# JavaScript SDK

The SDK provides a rich set of client-side functionality and is only about 14 kilobytes gzipped, which makes it ideal for mobile websites as well. It's a standalone SDK, which means it doesn't require external libraries like jQuery.

In this README:

- [Obtaining an API Key](#obtaining-an-api-key)
- [Loading and Initializing](#loading-and-initializing)
- [Geolocation](#geolocation)
- [API](#api)
- [Catalog View](#catalog-view)
- [Shopping List](#shopping-list)
- [File Upload](#file-upload)
- [Form Submit](#form-submit)

## Obtaining an API Key

To begin using the SDK, you must obtain an API key. Visit [our developer page](http://etilbudsavis.dk/developers/) where you can create an user and an app. Remember to enter the correct domain as the API uses CORS requiring the origin domain to be correct.

## Loading and Initializing

Place a `script` tag anywhere in your document to load the SDK synchronously (it goes without saying, you can load it asynchronously too):

```html
<script src="https://d2pnfhhphuc0n9.cloudfront.net/v2.1.0/eta.min.js"></script>
```

In development, we recommend you load the unminified source from `https://d2pnfhhphuc0n9.cloudfront.net/v2.1.0/eta.js` that logs to the console for easier debugging.

When the SDK has been loaded it's really easy to initialize it:

```js
var eta = new ETA.Client({
	apiKey: 'YOUR_API_KEY'
});
eta.initialize(function (err) {
	if (!err) {
		console.log('SDK has been initialized');
	} else {
		console.log('An error occured when initializing the SDK: ' + err);
	}
});
```

## Geolocation

If a geolocation has been set, the API will deliver results nearby that geolocation.

### Set a Geolocation

```js
eta.geolocation.set({
	latitude: 55.6309809,
	longitude: 12.5757921,
	radius: 400000, // Limits the results within a radius (optional)
	sensor: true // True if latitude/longitude was found using a sensor
});
```

All future API requests will use this geolocation until changed or reset.

### Get a Geolocation Property

```js
var latitude = eta.geolocation.get('latitude');
```

### Clone Geolocation Properties

```js
var attributes = eta.geolocation.clone();
```

### Reset a Geolocation

```js
eta.geolocation.reset();
```

### Calculate The Distance

To calculate the distance in metres from the currently set geolocation to another set of coordinates you simply do:

```js
var distance = eta.geolocation.getDistance(57.028811, 9.917771);
```

## API

Making API requests is very similar to when using jQuery AJAX. For example, to fetch a list of catalogs, you simply do:

```js
eta.api({
	url: '/v2/catalogs',
	success: function (data, statusText, xhr) {
		console.log('Catalogs fetched: ' + data);
	},
	error: function (xhr, statusText, errorThrown) {
		console.log('Catalogs not fetched: ' + statusText);
	}
});
```

To order the catalogs by popularity, that's really simple too:

```js
eta.api({
	url: '/v2/catalogs',
	data: {
		order_by: '-popularity'
	},
	success: function (data, statusText, xhr) {
		console.log('Catalogs fetched: ' + data);
	},
	error: function (xhr, statusText, errorThrown) {
		console.log('Catalogs not fetched: ' + statusText);
	}
});
```

Want to fetch for a specific dealer? You can use the `dealer_id` key. To learn more about the API, please see [our API documentation page](http://engineering.etilbudsavis.dk/eta-api/).

## Catalog View

We've made it dead simple to implement a catalog view on your own website:

```js
var catalogView = eta.catalogView();
catalogView.addEventListener('pagechange', function (e) {
	console.log('Pagechange: ' + e);
});
catalogView.open({
	catalog: 'A_CATALOG_ID'
});
```

That's it. Pretty simple, huh?

### Available Events

#### pagechange
Whenever a page changes.

#### outofbounds
When trying to go to a page that doesn't exists.

#### hotspot
When a hotspot is clicked.

#### singletap
When tapping a single time.

#### doubletap
When double tapping.

#### dragstart
When starting to drag a page.

#### analytics
Whenever a page changes and zooms (including a bunch of analytics data).

#### close
When the catalog view is closed.

## Shopping List

The shopping list is designed to either work on top of LocalStorage or our REST API. If you pass in an user id and your session allows acting on behalf of that user, all data changes will be synced to our backend. If no user is passed, LocalStorage will be used.

### Initialization

The shopping list constructor accepts two parameters: 1) options (optional) and 2) client (`ETA.Client` instance).

```js
var client = new ETA.Client({ apiKey: 'YOUR_API_KEY' });
var shoppingList = new ETA.ShoppingList(null, client);
```

To initialize the shopping list with an user, you simple do:

```js
var client = new ETA.Client({ apiKey: 'YOUR_API_KEY' });
var shoppingList = new ETA.ShoppingList({ userId: 'AN_USER_ID' }, client);
```

### Events

If you're familiar with event-based programming, you should feel right at home as the shopping list is entirely event-based.

An example of how one would integrate the shopping list with events is:

```js
shoppingList.addEventListener('list:add', function (list) {});
shoppingList.addEventListener('list:change', function (list) {});
shoppingList.addEventListener('list:remove', function (id) {});
shoppingList.addEventListener('list:empty', function (id) {});
shoppingList.addEventListener('lists:sync', function (removed, added, changed) {});
shoppingList.addEventListener('lists:reset', function () {});
shoppingList.addEventListener('item:add', function (item, list) {});
shoppingList.addEventListener('item:change', function (item, list) {});
shoppingList.addEventListener('item:remove', function (id, list) {});
shoppingList.addEventListener('items:sync', function (list, removed, added, changed) {});
shoppingList.addEventListener('items:reset', function (list) {});
shoppingList.addEventListener('share:add', function (list, email) {});
shoppingList.addEventListener('share:remove', function (list, email) {});
```

### Obtaining Lists

To obtain lists, you can either fetch them or get them from `localStorage`. An example would be:

```js
function obtainLists (callback) {
	if (shoppingList.hasUser) {
		// True if existing lists should be reset
		shoppingList.fetchLists(callback, true);
	} else {
		callback(null, shoppingList.getListsFromStorage());
	}
}

obtainLists(function (err, lists) {
	if (!err) {
		console.log('Obtained lists', lists);
	} else {
		console.log('Could not obtain lists', err);
	}
});
```

### Get List

You can easily get already obtained lists:

```js
var list = shoppingList.getList('LIST_ID');
```

### Add List

Adding lists is also easy:

```js
var list = shoppingList.addList({
	name: 'My shopping list',
	access: 'private' // Default value. Accepts `shared` as well
});
```

If you want to persist the shopping list you have to use `saveList` instead, which accepts the same arguments as `addList`.

To modify the list, simply pass the new list attributes to above methods as well.

### Remove List

If you wish to remove a list, you do:

```js
shoppingList.removeList('LIST_ID');
```

To persist the removal of the list, you have to do:

```js
shoppingList.destroyList('LIST_ID');
```

## File Upload

Uploading files cross-domain is a tricky business. We expose a class called `ETA.FileUpload`, which is a standalone file uploader that will return an identifier for the uploaded file. This identifier can then be used in endpoints that accept files instead of the actual file. This will also avoid having to upload the same file multiple times if you need it more than once.

The file uploader uses `FormData` if supported with an `iframe` fallback. The fallback needs you to host a blank HTML file on your site in order to work. Here's the contents of that HTML file:

	<!DOCTYPE HTML>
	<html>
		<head>
			<meta charset="utf-8">
		</head>
		<body></body>
	</html>

And this is how you use it:

```js
var file = document.getElementById('myFile');
var fileUpload = new ETA.FileUpload(file, 'http://mysite.com/redirect.html', client);
fileUpload.addEventListener('progress', function (e) {
	console.log('Current progress: ' + e);
});
fileUpload.send(function (err, id) {
	if (!err) {
		console.log('The file id is: ' + id);
	} else {
		console.log('Could not upload file: ' + err);
	}
});
```

## Form Submit
