(function () {
	var transports = {};
	var rheaders = /^(.*?):[ \t]*([^\r\n]*)$/mg;
	var rnoContent = /^(?:GET|HEAD)$/;
	var rQuery = /\?/;
	var allTypes = '*/'.concat('*');

	// For options that shouldn't be deep extended.
	var flatOptions = {
		url: true,
		context: true
	};

	// Data converters.
	var converters = {
		text: String,
		json: JSON.parse
	};

	// A special extend for ajax options that takes `flat` options (not to be deep extended).
	function ajaxExtend (target, src) {
		var key;
		var deep;

		for (key in src) {
			if (src[key] !== undefined) {
				(flatOptions[key] ? target : (deep || (deep = {})))[key] = src[key];
			}
		}
		if (deep) {
			ETA.utils.extend(true, target, deep);
		}

		return target;
	}

	// Convert response.
	function ajaxConvert (s, response, xhr, isSuccess) {
		var conv = converters[s.dataType];
		if (conv) {
			try {
				response = conv(response);
				if (s.responseFields[s.dataType]) {
					xhr[s.responseFields[s.dataType]] = response;
				}
			} catch (err) {
				return { state: 'parseerror', error: err };
			}
		}
		return { state: 'success', data: response };
	}

	// Get transport.
	function getTransport (s) {
		var transport = s.dataType.split(' ').shift();

		if (transports[transport]) {
			transport = transports[transport](s);
		} else {
			transport = undefined;
		}

		if (!transport) {
			for (var t in transports) {
				transport = transports[t](s);
				if (transport) break;
			}
		}

		return transport;
	}

	ETA.ajax = function (options) {
		// Force options to be an object.
		options = options || {};
		
		// Set defaults for options.
		var defaults = {
			type: 'get',
			dataType: 'json',
			processData: true,
			async: true,
			contentType: 'application/x-www-form-urlencoded; charset=UTF-8',
			timeout: 20000,
			responseFields: {
				xml: 'responseXML',
				text: 'responseText',
				json: 'responseJSON'
			},
			accepts: {
				'*': allTypes,
				text: 'text/plain',
				html: 'text/html',
				xml: 'application/xml, text/xml',
				json: 'application/json, text/javascript'
			}
		};

		// Create the final options object.
		var s = ajaxExtend(ajaxExtend({}, defaults), options);

		// Callbacks context.
		var callbackContext = s.context || s;

		// Transport.
		var transport;

		// Request/response headers.
		var requestHeaders = {};
		var responseHeaders;
		var responseHeadersString;

		// The XHR state.
		var state = 0;

		// Default abort message.
		var strAbort = 'canceled';

		// Timeout handle.
		var timeoutTimer;

		// Fake XHR.
		var xhr = {
			readyState: 0,
			getResponseHeader: function (key) {
				var match;
				if (state === 2) {
					if (!responseHeaders) {
						responseHeaders = {};
						while ((match = rheaders.exec(responseHeadersString))) {
							responseHeaders[match[1].toLowerCase()] = match[2];
						}
					}
					match = responseHeaders[key.toLowerCase()];
				}
				return match == null ? null : match;
			},
			getAllResponseHeaders: function () {
				return state === 2 ? responseHeadersString : null;
			},
			setRequestHeader: function (name, value) {
				if (!state) {
					requestHeaders[name] = value;
				}
				return this;
			},
			abort: function (statusText) {
				var finalText = statusText || strAbort;
				if (transport) {
					transport.abort(finalText);
				}
				done(0, finalText);
				return this;
			}
		};

		// Uppercase type.
		s.type = s.type.toUpperCase();

		// Convert data if not already a string.
		if (s.data && s.processData && typeof s.data !== 'string') {
			s.data = ETA.utils.param(s.data);
		}

		// Determine if request has content.
		s.hasContent = !rnoContent.test(s.type);

		// More options handling for requests with no content.
		if (!s.hasContent) {
			// If data is available, append data to URL.
			if (s.data) {
				s.url += (rQuery.test(s.url) ? '&' : '?') + s.data;
				delete s.data;
			}
		}

		// Set the correct header if data is being sent.
		if (s.data && s.contentType) {
			xhr.setRequestHeader('Content-Type', s.contentType);
		}

		// Set the Accepts header for the server, depending on the dataType.
		xhr.setRequestHeader(
			'Accept',
			s.dataType && s.accepts[s.dataType] ?
				s.accepts[s.dataType] + (s.dataType !== '*' ? ', ' + allTypes + '; q=0.01' : '') :
				s.accepts['*']
		);

		// Check for headers option.
		for (var header in s.headers) {
			xhr.setRequestHeader(header, s.headers[header]);
		}

		// Allow custom headers and early abort.
		if (s.beforeSend && (s.beforeSend.call(s, xhr) === false || state === 2)) {
			return xhr.abort();
		}

		// Aborting is no longer a cancellation.
		strAbort = 'abort';

		// Get transport.
		transport = getTransport(s);

		// If no transport, we auto-abort.
		if (!transport) {
			done(-1, 'No transport');
		} else {
			xhr.readyState = 1;

			// Timeout.
			if (s.async && s.timeout > 0) {
				timeoutTimer = setTimeout(function () {
					xhr.abort('timeout');
				}, s.timeout);
			}

			try {
				state = 1;
				transport.send(requestHeaders, done);
			} catch (err) {
				if (state < 2) {
					done(-1, err);
				} else {
					throw err;
				}
			}
		}

		// Callback for when everything is done.
		function done (status, nativeStatusText, response, headers) {
			var isSuccess;
			var success;
			var error;
			var statusText = nativeStatusText;

			// Called once.
			if (state === 2) {
				return;
			}

			// State is `done` now.
			state = 2;

			// Clear timeout if it exists.
			if (timeoutTimer) {
				clearTimeout(timeoutTimer);
			}

			// Deference transport.
			transport = undefined;

			// Cache response headers.
			responseHeadersString = headers || '';

			// Set readyState.
			xhr.readyState = status > 0 ? 4 : 0;

			// Determine if successful.
			isSuccess = status >= 200 && status < 300 || status === 304;

			// Convert no matter what.
			response = ajaxConvert(s, response, xhr, isSuccess);

			if (isSuccess) {
				// If no content.
				if (status === 204) {
					statusText = 'nocontent';
				
				// If not modified.
				} else if (status === 304) {
					statusText = 'notmodified';
				
				// If we have data, let's convert it.
				} else {
					statusText = response.state;
					success = response.data;
					error = response.error;
					isSuccess = !error;
				}
			} else {
				error = statusText;
				if (status || !statusText) {
					statusText = 'error';
					if (status < 0) {
						status = 0;
					}
				}
			}

			// Set data for the fake XHR object.
			xhr.status = status;
			xhr.statusText = statusText;

			// Success/error.
			if (isSuccess) {
				if (s.success) {
					s.success.call(callbackContext, success, statusText, xhr);
				}
			} else {
				if (s.error) {
					s.error.call(callbackContext, xhr, statusText, error);
				}
			}

			// Complete.
			if (s.complete) {
				s.complete.call(callbackContext, xhr, statusText);
			}
		}

		return xhr;
	};

	ETA.ajaxTransport = function (name, fn) {
		transports[name] = fn;
	};

	ETA.ajaxConverter = function (name, fn) {
		converters[name] = fn;
	};
})();
