// API.
//
// @param		Object			client			The client instance
// @param		Object			options			The request options
ETA.API = function (client, options) {
	this.client = client;
	this.options = options || {};
	this.sessionEnabled = options.session !== false;
	this.geolocationEnabled = options.geolocation !== false;
	this.aborted = false;
	this.errorCount = 0;
	this.maxErrors = 1;
	this.requestHeaders = {};
	this.makeRequest();
};

// Wrapper for making a request that ensures a session. The task queue ensures
// that multiple simultaneous requests are queued for session retrieval.
//
// @return		Object						This instance for chainability
ETA.API.prototype.makeRequest = function () {
	var self = this;
	if (this.sessionEnabled) {
		this.client.taskQueues.obtainSession.invoke(function (err) {
			self[!err ? '_makeRequest' : 'abort']();
		});
	} else {
		this._makeRequest();
	}
	return this;
};

// Private method for making a request after a session has been obtained.
//
// @return		void
ETA.API.prototype._makeRequest = function () {
	if (this.aborted || (this.sessionEnabled && !this.client.session.get('token'))) return;
	var options = this.options;
	var successCallback = options.success;
	var errorCallback = options.error;
	var completeCallback = options.complete;
	var data = {};
	var self = this;

	options.data = options.data || {};
	options.headers = options.headers || {};

	// Set token and signature request headers.
	if (this.sessionEnabled) {
		var sessionHeaders = this.client.session.getHeaders();
		for (var sessionHeader in sessionHeaders) {
			options.headers[sessionHeader] = sessionHeaders[sessionHeader];
		}
	}

	// Set remaining request headers.
	for (var requestHeader in this.requestHeaders) {
		options.headers[requestHeader] = this.requestHeaders[requestHeader];
	}

	// Override callbacks.
	options.success = function (data, statusText, xhr) {
		// Handle changes in session.
		if (self.sessionEnabled) {
			var currentToken = self.client.session.get('token');
			var returnedToken = xhr.getResponseHeader('X-Token');
			if (returnedToken && returnedToken !== currentToken) {
				self.client.session.set({ token: returnedToken }, true);
			}
		}

		if (successCallback) successCallback(data, statusText, xhr);
		if (completeCallback) completeCallback(xhr, statusText);
	};
	options.error = function (xhr, statusText, errorThrown) {
		var code = xhr.responseJSON && xhr.responseJSON.code;
		var retryAfter = +xhr.getResponseHeader('Retry-After');
		var done = function () {
			if (errorCallback) errorCallback(xhr, statusText, errorThrown);
			if (completeCallback) completeCallback(xhr, statusText);
		};

		// Expose API error.
		xhr.err = new Error('ApiError');
		xhr.err.name = self.errors[code] || 'UNKNOWN_ERROR';

		// Don't try to fix error if the error count has exceeded the
		// maximum errors allowed.
		if (self.errorCount >= self.maxErrors) {
			return done();
		}

		// Increment the error count.
		self.errorCount++;

		// Session errors.
		if (self.sessionEnabled && (code === 1101 || code === 1104 || code === 1108)) {
			self.client.taskQueues.createSession.invoke(function (err) {
				if (!err) {
					self._makeRequest();
				} else {
					done();
				}
			});

		// Temporary error - try again.
		} else if (code == 2015 && !isNaN(retryAfter) && retryAfter <= 1) {
			setTimeout(function () { self._makeRequest(); }, retryAfter * 1000);
		} else {
			done();
		}
	};
	options.complete = function (xhr, statusText) {};

	// Include geolocation if set and enabled.
	if (this.geolocationEnabled) {
		var latitude = this.client.geolocation.get('latitude');
		var longitude = this.client.geolocation.get('longitude');
		var radius = this.client.geolocation.get('radius');
		var sensor = this.client.geolocation.get('sensor');

		if ('r_lat' in options.data) {
			data.r_lat = options.data.r_lat;
		} else if (latitude != null) {
			data.r_lat = latitude;
		}
		if ('r_lng' in options.data) {
			data.r_lng = options.data.r_lng;
		} else if (longitude != null) {
			data.r_lng = longitude;
		}
		if ('r_radius' in options.data) {
			data.r_radius = options.data.r_radius;
		} else if (radius != null) {
			data.r_radius = radius;
		}
		if ('r_sensor' in options.data) {
			data.r_sensor = options.data.r_sensor;
		} else if (sensor != null) {
			data.r_sensor = sensor;
		}
	}

	// Apply parameters so that API-specific parameters come first in the query string.
	for (var key in options.data) data[key] = options.data[key];
	options.data = data;

	// Stringify `data` if content type is `application/json`.
	if (options.contentType === 'application/json' && options.data) {
		options.data = JSON.stringify(options.data);
	}

	// Prefix URL.
	if (options.url.substring(0, 1) === '/') {
		options.url = this.client.apiURL + options.url;
	}

	// Make the request.
	this.xhr = this.client.ajax(options);
};

// Set request header.
//
// @param		String			name			The name of the request header
// @param		String			value			The value of the request header
// @return		void
ETA.API.prototype.setRequestHeader = function (name, value) {
	this.requestHeaders[name] = value;
};

// Abort request.
//
// @return		void
ETA.API.prototype.abort = function () {
	this.aborted = true;
	
	if (this.xhr) {
		this.xhr.abort();
	} else {
		var xhr = this.getFakeXhr();
		if (this.options.error) {
			this.options.error(xhr, 'abort', 'abort');
		}
		if (this.options.complete) {
			this.options.complete(xhr, 'abort', 'abort');
		}
	}
};

// Get a fake XHR object.
//
// @return		Object							The fake XHR object
ETA.API.prototype.getFakeXhr = function () {
	return {
		readyState: 0,
		getResponseHeader: function () {},
		getAllResponseHeaders: function () {},
		setRequestHeader: function () {
			return this;
		},
		abort: function () {
			return this;
		}
	};
};

// Hash of errors we need to check for in unsuccessful requests.
ETA.API.prototype.errors = {
	// Session errors.
	1100: 'SESSION_ERROR',
	1101: 'SESSION_TOKEN_EXPIRED',
	1102: 'SESSION_INVALID_API_KEY',
	1103: 'SESSION_MISSING_SIGNATURE',
	1104: 'SESSION_INVALID_SIGNATURE',
	1105: 'SESSION_TOKEN_NOT_ALLOWED',
	1106: 'SESSION_MISSING_ORIGIN',
	1107: 'SESSION_MISSING_TOKEN',
	1108: 'SESSION_INVALID_TOKEN',

	// Authentication.
	1200: 'AUTHENTICATION_ERROR',
	1201: 'AUTHENTICATION_INVALID_CREDENTIALS',
	1202: 'AUTHENTICATION_NO_USER',
	1203: 'AUTHENTICATION_EMAIL_NOT_VERIFIED',

	// Authorization.
	1300: 'AUTHORIZATION_ERROR',
	1301: 'AUTHORIZATION_ACTION_NOT_ALLOWED',

	// Missing information.
	1400: 'INFO_REQUEST_INVALID',
	1401: 'INFO_MISSING_GEOLOCATION',
	1402: 'INFO_MISSING_RADIUS',
	1411: 'INFO_MISSING_AUTHENTICATION',
	1431: 'INFO_MISSING_EMAIL',
	1432: 'INFO_MISSING_BIRTHDAY',
	1433: 'INFO_MISSING_GENDER',
	1434: 'INFO_MISSING_LOCALE',
	1435: 'INFO_MISSING_NAME',
	1440: 'INFO_RESOURCE_NOT_FOUND',

	// Invalid information.
	1500: 'INFO_INVALID',
	1501: 'INFO_INVALID_RESOURCE_ID',
	1530: 'INFO_RESOURCE_DUPLICATION',
	1566: 'INFO_INVALID_BODY_DATA',

	// Internal corruption of data.
	2000: 'INTERNAL_INTEGRITY_ERROR',
	2010: 'INTERNAL_SEARCH_ERROR',
	2015: 'INTERNAL_NON_CRITICAL_ERROR',

	// Misc.
	4000: 'MISC_ACTION_NOT_EXISTS'
};
