// Main class to interface with.
//
// The configuration:
// - apiKey				The API key for requests.
// - apiSecret			The API secret for requests (optional).
// - apiURL				The main URL for API requests (optional).
// - apiPostMessageURL	The postMessage URL for API requests in browsers that don't support CORS (optional).
// - proxyURL			The proxy URL for communicating via the main website in an iframe (optional).
// - fileUploadURL		To support file uploads in Internet Explorer 8+9 (optional).
// - session			The session to use instead of looking in storage (optional).
// - geolocation		The default geolocation (optional).
// - locale				The preferred locale (optional).
// - storageEnabled		Set to false to disable storage (optional). Also takes an array to whitelist key prefixes.
// - tokenTTL			Decide the token time-to-live (TTL) when you create a new session (optional).
ETA.Client = function (options) {
	ETA.Dispatcher.mixin(this);
	
	// Ensure object.
	options = options || {};

	// You can retrieve an API key and secret at:
	// http://etilbudsavis.dk/developers/
	if (options.apiKey != null) {
		this.apiKey = options.apiKey;
	} else {
		var error = new Error('ClientInitializeError');
		error.name = 'API_KEY_MISSING';
		throw error;
	}

	if (options.apiSecret != null) {
		this.apiSecret = options.apiSecret;
	}

	if (options.apiURL != null) {
		this.apiURL = options.apiURL;
	}

	if (options.apiPostMessageURL != null) {
		this.apiPostMessageURL = options.apiPostMessageURL;
	}

	if (options.proxyURL != null) {
		this.proxyURL = options.proxyURL;
	}

	if (options.fileUploadURL != null) {
		this.fileUploadURL = options.fileUploadURL;
	}

	if (options.locale != null) {
		this.locale = options.locale;
	}

	if (options.storageEnabled != null) {
		this.storageEnabled = options.storageEnabled;
	}

	if (options.tokenTTL != null) {
		this.tokenTTL = options.tokenTTL;
	}

	this.apiPostMessageURL = this.apiPostMessageURL.replace('{api_key}', this.apiKey);
	this.usePostMessageTransport = !ETA.features.cors && ETA.features.postMessage;

	if (this.locale) {
		this.proxyURL = ETA.utils.updateParameter(this.proxyURL, 'locale', this.locale);
	}

	this.identifier = ETA.utils.uuid();

	this.setupSession(options.session);
	this.setupGeolocation(options.geolocation);
	this.setupTaskQueues();

	ETA.DOM.addEventListener('storage', window, ETA.utils.bind(this.storageHandle, this));

	if (DEBUG) {
		console.log('ETA.Client', 'Instantiated', options);
	}
};

ETA.Client.prototype.initialized = false;

ETA.Client.prototype.apiKey = null;
ETA.Client.prototype.apiSecret = null;

ETA.Client.prototype.mainURL = 'https://etilbudsavis.dk';
ETA.Client.prototype.apiURL = 'https://api.etilbudsavis.dk';
ETA.Client.prototype.apiPostMessageURL = 'https://api.etilbudsavis.dk/v2/proxy/{api_key}';
ETA.Client.prototype.proxyURL = 'https://etilbudsavis.dk/proxy/{id}/';
ETA.Client.prototype.fileUploadURL = location.protocol + '//' + location.hostname + ':' + location.port;

ETA.Client.prototype.locale = null;

ETA.Client.prototype.tokenTTL = undefined;

ETA.Client.prototype.storagePrefix = 'com.eta.sdk.';
ETA.Client.prototype.storageEnabled = true;

// Initialize the client.
//
// @return		void
ETA.Client.prototype.initialize = function (callback) {
	this.taskQueues.initialize.invoke(callback);
};

// Setup session.
//
// @param		Object			session			The predefined session to trump storage
// @return		void
ETA.Client.prototype.setupSession = function (session) {
	var self = this;
	var sessionsStorageKey = this.getStorageKey('sessions');
	var sessionStorageKey = this.getStorageKey('session');
	var sessionsStorage = this.getStorage(sessionsStorageKey);
	var sessionStorage = this.getStorage(sessionStorageKey);
	var sessions = sessionsStorage || {};

	if (ETA.utils.isObject(session)) {
		sessions[this.apiKey] = session;
	} else if (!sessionsStorage && ETA.utils.isObject(sessionStorage)) {
		sessions[this.apiKey] = sessionStorage;
		this.removeStorage(sessionStorageKey);
		this.setStorage(sessionsStorageKey, sessions);
	}

	this.session = new ETA.Session(sessions[this.apiKey], this);
	this.session.addEventListener('_change', function (attributes) {
		sessions[self.apiKey] = attributes;
		self.setStorage(sessionsStorageKey, sessions);
		self.dispatch('session:change', attributes);
	});
	this.session.addEventListener('_reset', function () {
		delete sessions[self.apiKey];
		self.setStorage(sessionsStorageKey, sessions);
	});
};

// Setup geolocation.
//
// @param		Object			geolocation		The predefined geolocation to trump storage
// @return		void
ETA.Client.prototype.setupGeolocation = function (geolocation) {
	var self = this;
	var geolocationStorageKey = this.getStorageKey('geolocation');

	this.geolocation = new ETA.Geolocation(geolocation || this.getStorage(geolocationStorageKey));
	this.geolocation.addEventListener('_change', function (attributes) {
		self.setStorage(geolocationStorageKey, attributes);
		self.dispatch('geolocation:change', attributes);
	});
	this.geolocation.addEventListener('_reset', function () {
		self.removeStorage(geolocationStorageKey);
	});
};

// Setup task queues.
//
// @return		void
ETA.Client.prototype.setupTaskQueues = function () {
	var self = this;

	this.taskQueues = {
		initialize: new ETA.TaskQueue(function (callback) {
			var token = self.session.get('token');

			if (self.initialized) {
				done();
			} else if (token) {
				renew();
			} else {
				create();
			}

			function done (err) {
				if (!err) self.initialized = true;
				callback(err);
			}

			function create () {
				self.session.create(done);
			}

			function renew () {
				self.session.renew(function (err) {
					if (!err) {
						done(err);
					} else {
						create();
					}
				});
			}
		}),
		obtainSession: new ETA.TaskQueue(function (callback) {
			self.session.obtain(callback);
		}),
		createSession: new ETA.TaskQueue(function (callback) {
			self.session.reset();
			self.session.create(callback);
		})
	};
};

// Make AJAX requests.
//
// @param		Object			options			The options
// @return		Object							The AJAX object
ETA.Client.prototype.ajax = function (options) {
	if (this.usePostMessageTransport) {
		options = options || {};
		options.data = options.data || {};
		options.data._ = new Date().getTime();
		options.dataType = 'postmessage ' + (options.dataType || 'json');
		options.postMessage = this.apiPostMessageURL;
		options.url = options.url.replace(this.apiURL, '');
		options.responseFields = {
			'postmessage xml': 'responseXML',
			'postmessage text': 'responseText',
			'postmessage json': 'responseJSON'
		};
	}
	
	return ETA.ajax(options);
};

// Make API requests.
//
// @param		Object			options			The options
// @return		Object							The API instance
ETA.Client.prototype.api = function (options) {
	return new ETA.API(this, options);
};

// Shopping list proxy.
//
// @param		Object			options			The options
// @return		Object							The exposed properties
ETA.Client.prototype.shoppingList = function (options) {
	var self = this;
	var proxy = new ETA.Proxy(ETA.utils.extend({}, {
		url: this.proxyURL,
		container: document.body
	}, options), this);
	var htmlElement = document.getElementsByTagName('html')[0];
	var originalOverflow = htmlElement.style.overflow;

	htmlElement.style.overflow = 'hidden';

	return {
		open: function (options) {
			proxy.initialize();
			proxy.send(['shopping-list', options]);

			setTimeout(function () {
				var contentWindow = proxy.iframe.contentWindow;
				if (contentWindow) contentWindow.focus();
			}, 100);
		},
		close: function () {
			htmlElement.style.overflow = originalOverflow;
			proxy.destroy();
		}
	};
};

// Catalog view proxy.
//
// @param		Object			options			The options
// @return		Object							The exposed properties
ETA.Client.prototype.catalogView = function (options) {
	var self = this;
	var catalogView = {};
	var proxy = new ETA.Proxy(ETA.utils.extend({}, {
		url: this.proxyURL,
		container: document.body
	}, options), this);
	var htmlElement = document.getElementsByTagName('html')[0];
	var originalOverflow = htmlElement.style.overflow;

	ETA.Dispatcher.mixin(catalogView);
	htmlElement.style.overflow = 'hidden';

	// Dispatch all catalog view events from the proxy to this vent.
	proxy.addEventListener('all', function (eventName, data) {
		if (/^eta-catalog-view-/i.test(eventName)) {
			catalogView.dispatch(eventName.replace('eta-catalog-view-', ''), data);
		}
	});

	// When the catalog is closed from within the proxy, the proxy has to be destroyed.
	catalogView.addEventListener('close', function () {
		htmlElement.style.overflow = originalOverflow;
		proxy.destroy();
	});

	return {
		addEventListener: function (eventName, callback) {
			catalogView.addEventListener(eventName, callback);
		},
		open: function (options) {
			proxy.initialize();
			proxy.send(['catalog-view', options]);

			setTimeout(function () {
				var contentWindow = proxy.iframe.contentWindow;
				if (contentWindow) contentWindow.focus();
			}, 100);
		},
		close: function () {
			proxy.send(['catalog-view-close']);
		},
		thumbnails: function () {
			proxy.send(['catalog-view-thumbnails']);
		},
		goToPage: function (page, animated) {
			proxy.send(['catalog-view-go-to-page', {
				page: page,
				animated: animated
			}]);
		}
	};
};

// Check if the storage key is allowed.
//
// @param		String			key				The storage key
// @return		Boolean							True if key is allowed
ETA.Client.prototype.allowStorageForKey = function (key) {
	var ok = false;
	var keyIndex;
	var keyLen;

	if (this.storageEnabled === true) {
		ok = true;
	} else if (ETA.utils.isArray(this.storageEnabled)) {
		keyIndex = 0;
		keyLen = this.storageEnabled.length;
		for (keyIndex; keyIndex < keyLen; keyIndex++) {
			if (key.indexOf(this.storageEnabled[keyIndex]) > -1) {
				ok = true;
				break;
			}
		}
	}

	return ok;
};

// Get storage key.
//
// @param		String			key				The key to prefix
// @return		String							The prefixed key
ETA.Client.prototype.getStorageKey = function (key) {
	return this.storagePrefix + key;
};

// Get data for key with a fallback to a default value.
//
// @param		String			key				The key
// @param		Mixed			defaultValue	The default value if nothing was found for key
// @return		Mixed							The found value
ETA.Client.prototype.getStorage = function (key, defaultValue) {
	if (this.allowStorageForKey(key)) {
		return ETA.Storage.get(key, defaultValue);
	} else {
		return defaultValue;
	}
};

// Set value for key with an optional lifetime.
//
// @param		String			key				The key
// @param		Mixed			data			The data
// @param		Number			lifetime		The lifetime in milliseconds
// @return		void
ETA.Client.prototype.setStorage = function (key, data, lifetime) {
	if (this.allowStorageForKey(key)) {
		ETA.Storage.set(key, data, lifetime);
	}
};

// Remove data for key.
//
// @param		String			key				The key
// @return		void
ETA.Client.prototype.removeStorage = function (key) {
	if (this.allowStorageForKey(key)) ETA.Storage.remove(key);
};

// Get all keys.
//
// @param		String			keyPrefix		Optional key prefix
// @return		Array							The saved keys
ETA.Client.prototype.getStorageKeys = function (keyPrefix) {
	if (this.allowStorageForKey(keyPrefix)) return ETA.Storage.getKeys(keyPrefix);
};

// When the storage changes in other windows.
//
// @param		Object			e				The event object
// @return		void
ETA.Client.prototype.storageHandle = function (e) {
	try {
		// The object event is stored in `window.event` in IE.
		if (!e) e = window.event;

		// Parse the value as JSON.
		var value = JSON.parse(e.newValue);
		
		// If this storage key was set by this window, stop execution.
		// The `storage` event should only be dispatched to other windows
		// but IE insists on dispatching the event in the same window too.
		if (value && value.id && ETA.Storage.ids[value.id]) return;

		// Check for session and geolocation keys as they need to be updated.
		if (e.key === this.getStorageKey('sessions')) {
			if (e.newValue != null) {
				this.session.attributes = JSON.parse(e.newValue).data[this.apiKey] || {};
				this.session.dispatch('change', this.session.clone());
				this.dispatch('session:change', this.session.clone());
			} else {
				this.session.attributes = {};
				this.session.dispatch('reset');
			}
		} else if (e.key === this.getStorageKey('geolocation')) {
			if (e.newValue != null) {
				this.geolocation.attributes = JSON.parse(e.newValue).data;
				this.geolocation.dispatch('change', this.geolocation.clone());
				this.dispatch('geolocation:change', this.geolocation.clone());
			} else {
				this.geolocation.attributes = {};
				this.geolocation.dispatch('reset');
			}
		}
	} catch (err) {}
};

// Get ERN URL.
//
// @param		String			ern				The ERN to use
// @return		String							The ERN URL
ETA.Client.prototype.getERNURL = function (ern) {
	var url = this.mainURL + '/ern/' + ern;
	if (this.locale) {
		url = ETA.utils.updateParameter(url, 'locale', this.locale);
	}
	return url;
};

// Redirect to main URL based on ERN.
//
// @param		String			ern				The ERN to use
// @return		void
ETA.Client.prototype.ERNRedirecter = function (ern) {
	location.href = this.getERNURL(ern);
};
