// Cross-domain and cross-browser file uploader.
//
// @param		Object			file			The input file field
// @param		Object			client			The client instance
ETA.FileUpload = function (file, client) {
	this.file = file;
	this.client = client;
	this.url = this.client.apiURL + '/v2/fileupload';
	this.type = 'POST';
	this.name = ETA.utils.uuid();
	this.useIframeTransport = !ETA.features.formData || !ETA.features.fileAPI || !ETA.features.cors;
};

ETA.Dispatcher.mixin(ETA.FileUpload);

// Send the file.
//
// @param		Function		callback		The callback
// @return		void
ETA.FileUpload.prototype.send = function (callback) {
	if (!this.file) {
		var error = new Error('FileUploadError');
		error.name = 'FILE_MISSING';
		throw error;
	}

	if (this.useIframeTransport) {
		this.sendUsingIframe(callback);
	} else {
		this.sendUsingFormData(callback);
	}
};

// Send the file using an iframe.
//
// @param		Function		callback		The callback
// @return		void
ETA.FileUpload.prototype.sendUsingIframe = function (callback) {
	var self = this;
	var fileClone;

	this.form = document.createElement('form');
	this.form.style.display = 'none';

	this.iframe = document.createElement('iframe');
	this.iframe.src = 'javascript'.concat(':false;');
	this.iframe.name = this.name;
	this.iframe.style.display = 'none';
	this.iframe.onload = function () {
		self.iframe.onload = function () {
			try {
				// Get parameter value.
				var response = ETA.utils.getParameter('_data', self.iframe.contentWindow.location.href);

				if (!response || response.length === 0) {
					var error = new Error('FileUploadError');
					error.name = 'IFRAME_RESPONSE_EMPTY';
					throw error;
				}

				// Decode base64 response.
				response = ETA.utils.base64.decode(response);

				// Parse as JSON.
				response = JSON.parse(response);

				// The complete callback returns the file identifier
				// to use in future requests.
				callback(null, response);
			} catch (err) {
				callback(err);
			}

			// Fix for IE endless progress bar activity bug
			// (happens on form submits to iframe targets):
			var ieIframe = document.createElement('iframe');
			ieIframe.src = 'javascript'.concat(':false;');
			self.form.appendChild(ieIframe);

			// Removing the form in a setTimeout call
			// allows Chrome's developer tools to display
			// the response result.
			setTimeout(function () {
				self.form.parentNode.removeChild(self.form);
				self.iframe.parentNode.removeChild(self.iframe);
			}, 0);
		};

		self.form.target = self.iframe.name;
		self.form.action = self.url;
		self.form.method = self.type;

		// Insert a clone for file input field.
		fileClone = self.file.cloneNode(true);
		self.file.parentNode.insertBefore(fileClone, self.file.nextSibling);

		// Append the URL as a hidden input field.
		var url = document.createElement('input');
		url.type = 'hidden';
		url.name = 'redirect';
		url.value = self.client.fileUploadURL;
		self.form.appendChild(url);

		// Appending the file input fields to the hidden form
		// removes them from their original location.
		self.file.name = 'file';
		self.form.appendChild(self.file);

		// `enctype` must be set as encoding for IE.
		self.form.enctype = 'multipart/form-data';
		self.form.encoding = 'multipart/form-data';

		// Submit the form.
		self.form.submit();

		// Insert the file input fields at their original location
		// by replacing the clones with the originals:
		self.file.name = fileClone.name;
		fileClone.parentNode.replaceChild(self.file, fileClone);
	};
	document.body.appendChild(this.iframe);
	this.iframe.appendChild(this.form);
};

// Send the file using form data.
//
// @param		Function		callback		The callback
// @return		void
ETA.FileUpload.prototype.sendUsingFormData = function (callback) {
	var self = this;
	var file = this.file.files[0];
	var formData = new FormData();
	formData.append('file', file);

	var xhr = this.xhr = new XMLHttpRequest();
	xhr.onload = function () {
		if (xhr.status === 200) {
			callback(null, JSON.parse(xhr.responseText));
		} else {
			var error = new Error('FileUploadError');
			error.name = 'XHR_SEND_ERROR';
			callback(error);
		}
	};
	xhr.onerror = function () {
		var error = new Error('FileUploadError');
		error.name = 'XHR_SEND_ERROR';
		callback(error);
	};
	xhr.upload.onprogress = function (e) {
		if (e.lengthComputable) {
			self.dispatch('progress', {
				progress: e.loaded / e.total,
				loaded: e.loaded,
				total: e.total
			});
		}
	};
	xhr.open(this.type, this.url, true);
	xhr.send(formData);
};

// Abort the request.
//
// @return		void
ETA.FileUpload.prototype.abort = function () {
	if (this.iframe) {
		// javascript:false as iframe src aborts the requests.
		// Concat is used to avoid the `Script URL` JSHint error.
		this.iframe.onload = function () {};
		this.iframe.src = 'javascript'.concat(':false;');
	}
	if (this.form) {
		this.form.parentNode.removeChild(this.form);
	}
	if (this.xhr) {
		this.xhr.abort();
	}
};
