// Form submitter with cross-domain and cross-browser file uploading capabilities.
//
// @param		Object			form			The form
// @param		Object			options			The options
// @param		Object			client			The client instance
ETA.FormSubmit = function (form, options, client) {
	this.form = form;
	this.options = options || {};
	this.options.url = this.options.url || this.form.action;
	this.options.type = this.options.type || this.form.method;
	this.client = client;
	this.fileDataAttribute = 'data-eta-file-upload-id';
};

// Submit the form.
//
// @return		void
ETA.FormSubmit.prototype.submit = function () {
	var self = this;
	var aborted = false;
	var api;

	this.uploadFiles(function () {
		if (aborted) return;
		var options = ETA.utils.extend({}, self.options);
		options.data = ETA.utils.extend(options.data, self.formToJSON());
		api = self.client.api(options);
	});

	return {
		abort: function () {
			aborted = true;

			if (api) {
				api.abort();
			} else {
				var xhr = ETA.API.prototype.getFakeXhr();
				if (self.options.error) {
					self.options.error(xhr, 'abort', 'abort');
				}
				if (self.options.complete) {
					self.options.complete(xhr, 'abort', 'abort');
				}
			}
		}
	};
};

// Upload files in form.
//
// @param		Function		callback		The callback
// @return		void
ETA.FormSubmit.prototype.uploadFiles = function (callback) {
	// Identify form elements that need uploading.
	var self = this;
	var files = [];
	var elements = this.form.elements;

	for (var i = 0, len = elements.length; i < len; i++) {
		var element = elements[i];
		if (!element.name || element.disabled || element.value.length === 0) continue;
		if (element.type === 'file') files.push(element);
	}

	// Keep uploading until no more files need to be uploaded.
	function upload () {
		if (files.length === 0) {
			callback();
			return;
		}

		// Start uploading the file first in row while removing
		// it from the stack.
		var file = files.shift();
		var fileUpload = new ETA.FileUpload(file, self.client);
		fileUpload.send(function (err, resp) {
			// Set data attribute on the input field to reference file id.
			if (!err) {
				file.setAttribute(self.fileDataAttribute, resp.file_id);
			}

			// Move on.
			upload();
		});
	}

	upload();
};

// Get the value of an input field.
//
// @param		Object			element			The input field
// @param		Boolean			successful		Controls whether or not the field element must be successful
// @return		String							The value
ETA.FormSubmit.prototype.fieldValue = function (element, successful) {
	var name = element.name;
	var type = element.type;
	var tag = element.tagName.toLowerCase();

	if (successful === undefined) successful = true;

	if (successful && (!name || element.disabled || type === 'reset' || type === 'button' ||
		(type === 'checkbox' || type === 'radio') && !element.checked ||
		tag === 'select' && element.selectedIndex === -1)) {
		return null;
	}

	if (tag === 'select') {
		var index = element.selectedIndex;
		if (index < 0) return null;
		var arr = [];
		var options = element.options;
		var one = type === 'select-one';
		var max = one ? index + 1 : options.length;

		for (var i = (one ? index : 0); i < max; i++) {
			var option = options[i];
			if (option.selected) {
				var value = option.value;
				if (!value) {
					value = (option.attributes && options.attributes.value && !(option.attributes.value.specified)) ? option.text : options.value;
				}
				if (one) return value;
				arr.push(v);
			}
		}
		return arr;
	}

	return element.value;
};

// Gather form element data into a JSON object.
//
// @return		Object								An object that holds all the form key:value pairs.
ETA.FormSubmit.prototype.formToJSON = function () {
	var data = {};
	var elements = this.form.elements;
	if (!elements) return data;

	for (var i = 0, len = elements.length; i < len; i++) {
		var element = elements[i];
		var name = element.name;
		if (!name || element.disabled) continue;

		var value = this.fieldValue(element);

		if (element.type === 'file') {
			var fileId = element.getAttribute(this.fileDataAttribute);
			if (fileId && fileId.length > 0) data[name] = fileId;
		} else if (value !== null && typeof value != 'undefined') {
			data[name] = value;
		}
	}

	return data;
};
