// Proxy for communicating in an iframe.
//
// @param		Object			options			The options
// @param		Object			client			The client instance
ETA.Proxy = function (options, client) {
	this.options = options || {};
	this.client = client;
	this.queue = [];
	this.isSupported = ETA.features.postMessage;
	this.id = ETA.utils.uuid();
	this.url = options.url.replace('{id}', this.id);
	this.target = ETA.utils.getTarget(this.url);
	this.hasLoaded = false;
	this.listenForEvents();
	this.createIframe();
};

ETA.Dispatcher.mixin(ETA.Proxy);

// Initialize proxy.
//
// @return		void
ETA.Proxy.prototype.initialize = function () {
	this.send(['initialize', {
		apiKey: this.client.apiKey,
		apiSecret: this.client.apiSecret,
		apiURL: this.client.apiURL,
		apiPostMessageURL: this.client.apiPostMessageURL,
		fileUploadURL: this.client.fileUploadURL,
		session: this.client.session.attributes,
		geolocation: this.client.geolocation.attributes
	}]);
};

// Run the message queue.
//
// @return		void
ETA.Proxy.prototype.runQueue = function () {
	if (!this.isSupported || !this.iframe || !this.hasLoaded || this.queue.length === 0) return;
	var data = this.queue.shift();
	var win = this.iframe.contentWindow || this.iframe.contentDocument.parentWindow;
	win.postMessage(JSON.stringify(data), this.target);
	this.runQueue();

	if (DEBUG) {
		console.log('ETA.Proxy', 'Message sent', data);
	}
};

// Create iframe used to proxy the messages.
//
// @return		void
ETA.Proxy.prototype.createIframe = function () {
	this.iframe = document.createElement('iframe');
	this.iframe.id = ['eta', 'proxy', this.id].join('-');
	this.iframe.className = 'eta-proxy';
	this.iframe.src = this.url;
	this.iframe.frameBorder = 0;
	this.iframe.allowTransparency = false;
	this.iframe.style.display = 'block';
	this.iframe.style.position = 'fixed';
	this.iframe.style.top = 0;
	this.iframe.style.left = 0;
	this.iframe.style.margin = 0;
	this.iframe.style.padding = 0;
	this.iframe.style.width = '100%';
	this.iframe.style.height = '100%';
	this.iframe.style.overflow = 'hidden';
	this.iframe.style.backgroundColor = '#ffffff';
	this.iframe.style.zIndex = 2147483647;

	this.options.container.appendChild(this.iframe);
};

// Events from the iframe comes via the window.
//
// @return		void
ETA.Proxy.prototype.listenForEvents = function () {
	var self = this;

	ETA.DOM.addEventListener('message', window, ETA.utils.bind(this.onMessage, this));

	// Notify the proxy if session changes.
	this.client.session.addEventListener('change', ETA.utils.bind(this.onSessionChange, this));

	// Notify the proxy if geolocation changes.
	this.client.geolocation.addEventListener('change', ETA.utils.bind(this.onGeolocationChange, this));

	// Update the session whenever it changes within the proxy.
	this.addEventListener('eta-session-change', function (attributes) {
		self.client.session.set(attributes, true);
	});

	// Update the geolocation whenever it changes within the proxy.
	this.addEventListener('eta-geolocation-change', function (attributes) {
		self.client.geolocation.set(attributes, true);
	});
};

// Send data to the iframe.
//
// @param		Object			data			The data to send
// @return		void
ETA.Proxy.prototype.send = function (data) {
	this.queue.push(data);
	this.runQueue();
};

// Destroy the proxy.
//
// @return		void
ETA.Proxy.prototype.destroy = function () {
	ETA.DOM.removeEventListener('message', window, this.onMessage);
	this.iframe.parentNode.removeChild(this.iframe);

	this.client.removeEventListener('session:change', this.onSessionChange);
	this.client.removeEventListener('geolocation:change', this.onGeolocationChange);

	if (DEBUG) {
		console.log('ETA.Proxy', 'Destroyed');
	}
};

// When messages are received from the window.
//
// @param		Object			e				The event data
// @return		void
ETA.Proxy.prototype.onMessage = function (e) {
	try {
		var data = JSON.parse(e.data);
		if (data && data.id === this.id) {
			if (data.eventName === 'eta-proxy-ready') {
				this.hasLoaded = true;
				this.runQueue();
			}
			this.dispatch(data.eventName, data.data);
		}

		if (DEBUG) {
			console.log('ETA.Proxy', 'Message received', data, e.origin);
		}
	} catch (err) {
		if (DEBUG) {
			console.log('ETA.Proxy', 'Could not parse message', e.origin);
		}
	}
};

// When the session changes.
//
// @param		Object			attributes		The session attributes
// @return		void
ETA.Proxy.prototype.onSessionChange = function (attributes) {
	this.send(['session-change', attributes]);
};

// When the geolocation changes.
//
// @param		Object			attributes		The geolocation attributes
// @return		void
ETA.Proxy.prototype.onGeolocationChange = function (attributes) {
	this.send(['geolocation-change', attributes]);
};
