// Session management.
//
// @param		Object			attributes		The attributes
// @param		Object			client			The client instance
ETA.Session = function (attributes, client) {
	this.attributes = attributes || {};
	this.client = client;
	this.url = this.client.apiURL + '/v2/sessions';
};

ETA.Dispatcher.mixin(ETA.Session);

// Get attribute key value.
//
// @param		String			key				The key
// @return		Mixed							The attribute key value
ETA.Session.prototype.get = function (key) {
	return this.attributes[key];
};

// Set attributes.
//
// @param		Object			attributes		The session attributes
// @param		Boolean			silent			True if no events should be propagated
// @return		void
ETA.Session.prototype.set = function (attributes, silent) {
	var changes = false;
	for (var key in attributes) {
		if (this.attributes[key] !== attributes[key]) {
			this.attributes[key] = attributes[key];
			changes = true;
		}
	}
	if (changes) {
		var clone = this.clone();
		this.dispatch('_change', clone);
		if (!silent) this.dispatch('change', clone);

		if (DEBUG) {
			console.log('ETA.Session', 'Changed', clone);
		}
	}
};

// Get request headers needed for API requests.
//
// @return		Object							The headers
ETA.Session.prototype.getHeaders = function () {
	var headers = {};
	var token = this.get('token');
	if (token) {
		headers['X-Token'] = token;
		if (this.client.apiSecret) headers['X-Signature'] = this.getSignature();
	}
	return headers;
};

// Get signature.
//
// @return		String							The signature
ETA.Session.prototype.getSignature = function () {
	return ETA.utils.SHA256([this.client.apiSecret, this.get('token')].join(''));
};

// Has session expired?
//
// @return		Boolean							True if session has expired
ETA.Session.prototype.hasExpired = function () {
	if (this.get('token') && this.get('expires')) {
		return new Date().getTime() >= ETA.utils.parseISOString(this.get('expires'));
	} else {
		return true;
	}
};

// Has an user been attached to the session?
//
// @return		Boolean							True if an user has been attached
ETA.Session.prototype.isUserAttached = function () {
	return !!this.get('user');
};

// Get the user id attached to the session.
//
// @return		Number							The user id
ETA.Session.prototype.getUserId = function () {
	if (this.isUserAttached()) return this.get('user').id;
};

// Clone session attributes.
//
// @return		Object							The session attributes
ETA.Session.prototype.clone = function () {
	return ETA.utils.clone(this.attributes);
};

// Reset attributes.
//
// @param		Boolean			silent			True if no events should be propagated
// @return		void
ETA.Session.prototype.reset = function (silent) {
	this.attributes = {};
	this.dispatch('_reset');
	if (!silent) this.dispatch('reset');
};

// Check if user is allowed.
//
// @param		Array			actions			Required permission actions
// @return		Boolean							True if user is allowed
ETA.Session.prototype.isUserAllowed = function (actions) {
	var user = this.get('user');
	if (!user) return false;
	if (!actions || actions.length === 0) return true;
	return this.allowsAll(actions);
};

// Check if permissions stored in session allows given action.
//
// @param		String			action			The action
// @return		Boolean							True if action is allowed
ETA.Session.prototype.allows = function (action) {
	// Don't allow if no permissions.
	var permissions = this.get('permissions');
	if (!permissions) return false;

	for (var name in permissions) {
		var permissionCollection = new ETA.Permissions(permissions[name], name);
		if (permissionCollection.allows(action)) {
			return true;
		}
	}
	return false;
};

// Check if permissions stored in session allows all given actions.
//
// @param		Array			actions			The actions
// @return		Boolean							True if actions are allowed
ETA.Session.prototype.allowsAll = function (actions) {
	for (var i = 0, len = actions.length; i < len; i++) {
		if (!this.allows(actions[i])) {
			return false;
		}
	}
	return true;
};

// Obtain session attributes.
//
// @param		Function		callback		The callback
// @return		void
ETA.Session.prototype.obtain = function (callback) {
	var token = this.get('token');
	var expires = this.get('expires');
	var now = new Date().getTime();
	var day = 1000 * 60 * 60 * 24 * 2;
	var self = this;

	function done (err, isNew) {
		if (!err) {
			callback(null, self.clone(), isNew);
		} else {
			callback(err);
		}
	}

	if (token && expires) {
		if (now + day <= ETA.utils.parseISOString(expires)) {
			done(null, false);
		} else {
			this.renew(function (err) {
				if (!err) {
					done(err, true);
				} else {
					self.create(function (err) {
						done(err, true);
					});
				}
			});
		}
	} else {
		this.create(function (err) {
			done(err, true);
		});
	}
};

// Create session via the API.
//
// @param		Function		callback		The callback
// @return		void
ETA.Session.prototype.create = function (callback) {
	var self = this;
	this.client.api({
		type: 'post',
		url: this.url,
		data: {
			api_key: this.client.apiKey,
			token_ttl: this.client.tokenTTL
		},
		session: false,
		geolocation: false,
		success: function (data) {
			self.set(data);
			callback(null, self.clone());
		},
		error: function (xhr) {
			callback(xhr.err);
		}
	});
};

// Update session via the API.
//
// @param		Function		callback		The callback
// @return		void
ETA.Session.prototype.update = function (callback) {
	var self = this;
	var error;

	if (!this.get('token')) {
		error = new TypeError('SessionError');
		error.name = 'TOKEN_MISSING';
		return callback(error);
	}

	this.client.api({
		url: this.url,
		headers: this.getHeaders(),
		session: false,
		geolocation: false,
		success: function (data) {
			self.set(data);
			callback(null, self.clone());
		},
		error: function (xhr) {
			callback(xhr.err);
		}
	});
};

// Renew session via the API.
//
// @param		Function		callback		The callback
// @return		void
ETA.Session.prototype.renew = function (callback) {
	var self = this;
	var error;

	if (!this.get('token')) {
		error = new TypeError('SessionError');
		error.name = 'TOKEN_MISSING';
		return callback(error);
	}

	this.client.api({
		type: 'put',
		url: this.url,
		headers: this.getHeaders(),
		session: false,
		geolocation: false,
		success: function (data) {
			self.set(data);
			callback(null, self.clone());
		},
		error: function (xhr) {
			callback(xhr.err);
		}
	});
};

// Attach user to session via the API.
//
// @param		Object			credentials		The user credentials
// @param		Function		callback		The callback
// @return		void
ETA.Session.prototype.attachUser = function (credentials, callback) {
	var self = this;

	this.detachUser(function () {
		self.client.api({
			type: 'put',
			url: self.url,
			data: credentials,
			success: function (data) {
				self.set(data);
				callback(null, self.clone());
			},
			error: function (xhr) {
				callback(xhr.err);
			}
		});
	});
};

// Detach user from session via the API.
//
// @param		Function		callback		The callback
// @return		void
ETA.Session.prototype.detachUser = function (callback) {
	var self = this;

	this.client.api({
		type: 'put',
		url: this.url,
		data: {
			email: ''
		},
		success: function (data) {
			self.attributes = {};
			self.set(data);
			callback(null, self.clone());
		},
		error: function (xhr) {
			callback(xhr.err);
		}
	});
};
