// Shopping list base.
//
// @param		Object			options			The options
// @param		Object			client			The client
ETA.ShoppingList = function (options, client) {
	// Ensure object.
	options = options || {};

	if (options.userId != null) {
		this.updateUser(options.userId);
	}

	if (options.listsPollDelay != null) {
		this.listsPollDelay = options.listsPollDelay;
	}
	
	if (options.itemsPollDelay != null) {
		this.itemsPollDelay = options.itemsPollDelay;
	}

	this.client = client;
	this.lists = [];
	this.items = {};
	this.offers = {};
	this.stores = {};

	this.apiRequests = [];
	this.apiRunning = false;

	// Prepopulate lists and items from storage when there's no user.
	if (!this.hasUser) {
		this.lists = this.getListsFromStorage();
		this.items = this.getItemsFromStorage();
	}
};

ETA.Dispatcher.mixin(ETA.ShoppingList);

ETA.ShoppingList.prototype.userId = null;
ETA.ShoppingList.prototype.hasUser = false;

ETA.ShoppingList.prototype.storagePrefix = 'com.eta.sdk.shopping_list.';

ETA.ShoppingList.prototype.baseURL = '/v2/users/{user_id}/shoppinglists';

ETA.ShoppingList.prototype.listsPollDelay = 20 * 1000;
ETA.ShoppingList.prototype.itemsPollDelay = 6 * 1000;
ETA.ShoppingList.prototype.listsPollTimeout = null;
ETA.ShoppingList.prototype.itemsPollTimeout = null;

// List
// -------------

// Get lists from storage.
//
// @return		Object							The lists
ETA.ShoppingList.prototype.getListsFromStorage = function () {
	return this.client.getStorage(this.storagePrefix + 'lists', []);
};

// Fetch lists.
//
// @param		Function		callback		The callback
// @param		Boolean			reset			True if existing lists should be reset
// @return		void
ETA.ShoppingList.prototype.fetchLists = function (callback, reset) {
	var self = this;
	var error;

	if (!this.hasUser) return;

	this.api({
		url: this.getBaseURL(),
		success: function (lists) {
			self[!reset ? 'synchronizeLists' : 'resetLists'](lists);
			callback(null, self.lists);
		},
		error: function (xhr) {
			callback(xhr.err);
		}
	});
};

// Fetch list.
//
// @param		String			id				The id of the list
// @param		Function		callback		The callback
// @return		void
ETA.ShoppingList.prototype.fetchList = function (id, callback) {
	var self = this;
	var error;

	if (!this.hasUser) return;

	this.api({
		url: [this.getBaseURL(), id].join('/'),
		success: function (list) {
			callback(null, list);
		},
		error: function (xhr) {
			callback(xhr.err);
		}
	});
};

// Get list.
//
// @param		String			id				The id of the list
// @return		Object							The list
ETA.ShoppingList.prototype.getList = function (id) {
	for (var i = 0, len = this.lists.length; i < len; i++) {
		var list = this.lists[i];
		if (list.id === id) return list;
	}
};

// Save list.
//
// @param		Object			attributes		The list attributes
// @param		Boolean			preconstructed	True if the attributes shouldn't be modified
// @param		Boolean			silent			True if no event should be dispatched
// @return		Object							The new list
ETA.ShoppingList.prototype.saveList = function (attributes, preconstructed, silent) {
	var api = this.apiPromise();
	var list = this.setList(attributes, preconstructed, silent);
	if (!this.hasUser) {
		this.client.setStorage(this.storagePrefix + 'lists', this.lists);
		api.reject();
	} else {
		api.resolve({
			type: 'put',
			url: [this.getBaseURL(), list.id].join('/'),
			data: list
		});
	}
	return list;
};

// Add list (wrapper for `setList`).
ETA.ShoppingList.prototype.addList = function () {
	return this.setList.apply(this, arguments);
};

// Set list attributes.
//
// @param		Object			attributes		The list attributes
// @param		Boolean			preconstructed	True if the attributes shouldn't be modified
// @param		Boolean			silent			True if no event should be dispatched
// @return		Object							The new list
ETA.ShoppingList.prototype.setList = function (attributes, preconstructed, silent) {
	attributes = attributes || {};
	var list = this.getList(attributes.id);
	var isNew = !list;

	if (!preconstructed) {
		if (isNew) {
			list = {};
			list.id = ETA.utils.uuid();
		}

		for (var defaultKey in this.listDefaults) {
			if (list[defaultKey] == null) {
				list[defaultKey] = this.listDefaults[defaultKey];
			}
		}

		for (var key in attributes) {
			list[key] = attributes[key];
		}

		list.modified = this.getModified();
	} else {
		list = attributes;
	}

	if (isNew) {
		this.lists.push(list);
		if (!silent) this.dispatch('list:add', list);
	} else {
		if (!silent) this.dispatch('list:change', list);
	}

	return list;
};

// Remove list.
//
// @param		String			id				The list
// @param		Boolean			silent			True if no event should be dispatched
// @return		void
ETA.ShoppingList.prototype.removeList = function (id, silent) {
	for (var i = 0, len = this.lists.length; i < len; i++) {
		if (this.lists[i].id === id) {
			this.lists.splice(i, 1);
			if (!silent) this.dispatch('list:remove', id);
			break;
		}
	}
	delete this.items[id];
};

// Destroy list.
//
// @param		String			id				The list
// @return		void
ETA.ShoppingList.prototype.destroyList = function (id) {
	this.removeList(id);
	if (!this.hasUser) {
		this.client.setStorage(this.storagePrefix + 'lists', this.lists);
		this.client.removeStorage(this.storagePrefix + 'items.' + id);
	} else {
		this.api({
			type: 'delete',
			url: [this.getBaseURL(), id].join('/'),
			data: {
				modified: this.getModified()
			}
		});
	}
};

// Has a list been modified?
//
// @param		String			id				The list
// @param		Function		callback		The callback
// @return		void
ETA.ShoppingList.prototype.hasListBeenModified = function (id, callback) {
	var list = this.getList(id);
	if (!list) {
		callback(null, false);
	} else {
		var lastModified = ETA.utils.parseISOString(list.modified);
		this.api({
			url: [this.getBaseURL(), id, 'modified'].join('/'),
			success: function (resp) {
				var modified = ETA.utils.parseISOString(resp.modified);
				var hasBeenModified = modified > lastModified;
				if (hasBeenModified) list.modified = resp.modified;
				callback(null, hasBeenModified);

				if (DEBUG) {
					if (hasBeenModified) {
						console.log('ETA.ShoppingList', 'Has been modified', list);
					}
				}
			},
			error: function (xhr) {
				callback(xhr.err);
			}
		});
	}
};

// Synchronize lists.
//
// @param		Array			lists			The lists
// @param		Boolean			silent			True if no event should be dispatched
// @return		void
ETA.ShoppingList.prototype.synchronizeLists = function (lists, silent) {
	var removed = [];
	var added = [];
	var changed = [];

	// Check for lists to be removed.
	for (var a = 0; a < this.lists.length; a++) {
		var exists = false;
		for (var b = 0; b < lists.length; b++) {
			if (this.lists[a].id === lists[b].id) {
				exists = true;
				break;
			}
		}

		if (!exists) removed.push(this.lists[a]);
	}

	// Check for lists to be added or changed.
	for (var c = 0; c < lists.length; c++) {
		var currentList = this.getList(lists[c].id);
		if (!currentList) {
			added.push(lists[c]);
		} else {
			var modified = ETA.utils.parseISOString(lists[c].modified);
			var currentModified = ETA.utils.parseISOString(currentList.modified);
			if (modified > currentModified) {
				changed.push(lists[c]);
			}
		}
	}

	for (var d = 0; d < removed.length; d++) {
		this.removeList(removed[d].id, true);
	}

	for (var e = 0; e < added.length; e++) {
		this.setList(added[e], true, true);
	}

	for (var f = 0; f < changed.length; f++) {
		this.setList(changed[f], true, true);
	}

	if (!silent) this.dispatch('lists:sync', removed, added, changed);
};

// Reset existing lists.
//
// @param		Array			lists			The lists
// @param		Boolean			silent			True if no event should be dispatched
// @return		Array							The same lists
ETA.ShoppingList.prototype.resetLists = function (lists, silent) {
	this.lists = lists || [];
	if (!silent) this.dispatch('lists:reset');
	return this.lists;
};

// Empty list.
//
// @param		String			id				The list
// @param		String			filter			Filter to empty items by
// @return		void
ETA.ShoppingList.prototype.emptyList = function (id, filter) {
	if (!filter || filter === 'all') {
		this.resetItems(null, id, true);
	} else {
		var filteredItems = [];
		var items = this.getItems(id);
		for (var i = 0, len = items.length; i < len; i++) {
			var item = items[i];
			if (filter === 'ticked' && !item.tick) {
				filteredItems.push(item);
			} else if (filter === 'unticked' && item.tick) {
				filteredItems.push(item);
			}
		}
		this.items[id] = filteredItems;
	}
	
	if (!this.hasUser) {
		this.client.setStorage(this.storagePrefix + 'items.' + id, this.getItems(id));
	} else {
		var self = this;
		this.api({
			type: 'post',
			url: [this.getBaseURL(), id, 'empty'].join('/'),
			data: {
				filter: filter,
				modified: this.getModified()
			},
			error: function (xhr) {
				if (xhr.status === 404) self.removeList(id);
			}
		});
	}

	this.dispatch('list:empty', id);
};

// Items
// -------------

// Get items from storage.
//
// @return		Object							The items
ETA.ShoppingList.prototype.getItemsFromStorage = function () {
	var items = {};
	var keys = this.client.getStorageKeys(this.storagePrefix + 'items.') || [];
	for (var i = 0; i < keys.length; i++) {
		var tokens = keys[i].split('.');
		var list = tokens[tokens.length - 1];
		items[list] = this.client.getStorage(keys[i]);
	}
	return items;
};

// Fetch items.
//
// @param		String			list			The list
// @param		Function		callback		The callback
// @param		Boolean			reset			True if existing items should be reset
// @return		void
ETA.ShoppingList.prototype.fetchItems = function (list, callback, reset) {
	if (!this.hasUser) return;
	var self = this;
	this.api({
		url: [this.getBaseURL(), list, 'items'].join('/'),
		success: function (items) {
			self.fetchRelationsForItems(items, function (err) {
				if (!err) {
					self[!reset ? 'synchronizeItems' : 'resetItems'](items, list);
					callback(null, self.getItems(list));
				} else {
					callback(err);
				}
			});
		},
		error: function (xhr) {
			if (xhr.status === 404) self.removeList(list);
			callback(xhr.err);
		}
	});
};

// Fetch relations for items.
//
// @param		Array			items			The items to fetch for
// @param		Function		callback		The callback
// @return		void
ETA.ShoppingList.prototype.fetchRelationsForItems = function (items, callback) {
	var self = this;
	var offers = [];
	var stores = [];

	items = items || [];

	for (var i = 0; i < items.length; i++) {
		var offerId = items[i].offer_id;
		if (offerId && !this.getOffer(offerId)) {
			offers.push(offerId);
		}
	}

	if (offers.length === 0) {
		callback(null);
		return;
	}

	this.api({
		url: '/v2/offers',
		data: {
			offer_ids: offers.join(',')
		},
		success: function (offers) {
			for (var k = 0; k < offers.length; k++) {
				var storeId = offers[k].store_id;
				if (storeId && !self.getStore(storeId)) {
					stores.push(storeId);
				}
				self.offers[offers[k].id] = offers[k];
			}

			if (stores.length === 0) {
				callback(null);
				return;
			}

			self.api({
				url: '/v2/stores',
				data: {
					store_ids: stores.join(',')
				},
				success: function (stores) {
					for (var j = 0; j < stores.length; j++) {
						self.stores[stores[j].id] = stores[j];
					}
					callback(null);
				},
				error: function (xhr) {
					callback(xhr.err);
				}
			});
		},
		error: function (xhr) {
			callback(xhr.err);
		}
	});
};

// Get items.
//
// @param		String			list			The id of the list
// @return		Array							The items
ETA.ShoppingList.prototype.getItems = function (list) {
	return this.items[list];
};

// Get item from list.
//
// @param		String			id				The item id
// @param		String			list			The list id
// @return		Object							The item attributes
ETA.ShoppingList.prototype.getItemFromList = function (id, list) {
	var items = this.getItems(list);
	if (!items) return;
	for (var i = 0, len = items.length; i < len; i++) {
		if (items[i].id === id) return items[i];
	}
};

// Save item.
//
// @param		Object			attributes		The item attributes
// @param		String			list			The list
// @param		Boolean			preconstructed	True if the attributes shouldn't be modified
// @param		Boolean			silent			True if no event should be dispatched
// @return		Objet							The item
ETA.ShoppingList.prototype.saveItem = function (attributes, list, preconstructed, silent) {
	var api = this.apiPromise();
	var item = this.setItem(attributes, list, preconstructed, silent);
	if (!this.hasUser) {
		this.client.setStorage(this.storagePrefix + 'items.' + list, this.getItems(list));
		api.reject();
	} else {
		api.resolve({
			type: 'put',
			url: [this.getBaseURL(), list, 'items', item.id].join('/'),
			data: item
		});
	}
	return item;
};

// Add item (wrapper for `setItem`).
ETA.ShoppingList.prototype.addItem = function () {
	return this.setItem.apply(this, arguments);
};

// Set item attributes.
//
// @param		Object			attributes		The item attributes
// @param		String			list			The list
// @param		Boolean			preconstructed	True if the attributes shouldn't be modified
// @param		Boolean			silent			True if no event should be dispatched
// @return		Object							The new item
ETA.ShoppingList.prototype.setItem = function (attributes, list, preconstructed, silent) {
	attributes = attributes || {};
	var item = this.getItemFromList(attributes.id, list);
	var isNew = !item;

	if (!preconstructed) {
		if (isNew) {
			item = {};
			item.id = ETA.utils.uuid();
		}

		for (var defaultKey in this.itemDefaults) {
			if (item[defaultKey] == null) {
				item[defaultKey] = this.itemDefaults[defaultKey];
			}
		}

		for (var key in attributes) {
			item[key] = attributes[key];
		}

		item.modified = this.getModified();
	} else {
		item = attributes;
	}

	if (isNew) {
		if (!this.items[list]) this.items[list] = [];
		this.items[list].push(item);
		if (!silent) this.dispatch('item:add', item, list);
	} else {
		if (!silent) this.dispatch('item:change', item, list);
	}

	return item;
};

// Remove item.
//
// @param		String			id				The item
// @param		String			list			The list
// @param		Boolean			silent			True if no event should be dispatched
// @return		void
ETA.ShoppingList.prototype.removeItem = function (id, list, silent) {
	var items = this.getItems(list) || [];
	var index;
	for (var i = 0, len = items.length; i < len; i++) {
		if (items[i].id === id) {
			index = i;
			break;
		}
	}
	if (index != null) {
		items.splice(index, 1);
		if (!silent) this.dispatch('item:remove', id, list);
	}
};

// Destroy item.
//
// @param		String			id				The item
// @param		String			list			The list
// @return		void
ETA.ShoppingList.prototype.destroyItem = function (id, list) {
	this.removeItem(id, list);
	if (!this.hasUser) {
		var items = this.getItems(list) || [];
		if (items.length > 0) {
			this.client.setStorage(this.storagePrefix + 'items.' + list, items);
		} else {
			this.client.removeStorage(this.storagePrefix + 'items.' + list);
		}
	} else {
		var self = this;
		this.api({
			type: 'delete',
			url: [this.getBaseURL(), list, 'items', id].join('/'),
			data: {
				modified: this.getModified()
			},
			error: function (xhr) {
				if (xhr.status === 404) self.removeList(list);
			}
		});
	}
};

// Synchronize items triggering the proper events.
//
// @param		Array			items			The items
// @param		String			list			The list
// @param		Boolean			silent			True if no event should be dispatched
// @return		void
ETA.ShoppingList.prototype.synchronizeItems = function (items, list, silent) {
	var removed = [];
	var added = [];
	var changed = [];

	// Check for items to be removed.
	var currentItems = this.getItems(list) || [];
	for (var a = 0; a < currentItems.length; a++) {
		var exists = false;
		for (var b = 0; b < items.length; b++) {
			if (currentItems[a].id === items[b].id) {
				exists = true;
				break;
			}
		}

		if (!exists) removed.push(currentItems[a]);
	}

	// Check for items to be added or changed.
	for (var c = 0; c < items.length; c++) {
		var currentItem = this.getItemFromList(items[c].id, list);
		if (!currentItem) {
			added.push(items[c]);
		} else {
			var modified = ETA.utils.parseISOString(items[c].modified);
			var currentModified = ETA.utils.parseISOString(currentItem.modified);
			if (modified > currentModified) {
				changed.push(items[c]);
			}
		}
	}

	for (var d = 0; d < removed.length; d++) {
		this.removeItem(removed[d].id, list, true);
	}

	for (var e = 0; e < added.length; e++) {
		this.setItem(added[e], list, true, true);
	}

	for (var f = 0; f < changed.length; f++) {
		this.setItem(changed[f], list, true, true);
	}

	if (!silent) this.dispatch('items:sync', list, removed, added, changed);
};

// Reset existing items for a list.
//
// @param		Array			items			The items
// @param		String			list			The list
// @param		Boolean			silent			True if no event should be dispatched
// @return		Array							The same items
ETA.ShoppingList.prototype.resetItems = function (items, list, silent) {
	this.items[list] = items || [];
	if (!silent) this.dispatch('items:reset', list);
	return this.items[list];
};

// Reset existing items for all lists.
//
// @param		Object			items			The items
// @param		Boolean			silent			True if no event should be dispatched
// @return		Object							The same items
ETA.ShoppingList.prototype.resetAllItems = function (items, silent) {
	this.items = items || {};
	if (!silent) this.dispatch('items:reset');
	return this.items;
};

// Shares
// -------------

// Fetch shares for list.
//
// @param		String			list			The list
// @param		Function		callback		The callback
// @return		void
ETA.ShoppingList.prototype.fetchShares = function (list, callback) {
	if (!this.hasUser) return;
	var self = this;
	this.api({
		url: [this.getBaseURL(), list, 'shares'].join('/'),
		success: function (shares) {
			callback(null, shares);
		},
		error: function (xhr) {
			if (xhr.status === 404) self.removeList(list);
			callback(xhr.err);
		}
	});
};

// Save share for list.
//
// @param		String			list			The list
// @param		String			email			The email
// @param		Object			data			The data
// @param		Function		callback		The callback
// @param		Boolean			silent			True if no event should be dispatched
// @return		void
ETA.ShoppingList.prototype.saveShare = function (list, email, data, callback, silent) {
	if (!this.hasUser) return;
	var self = this;
	this.api({
		type: 'put',
		url: [this.getBaseURL(), list, 'shares', encodeURIComponent(email)].join('/'),
		data: data,
		success: function (share) {
			callback(null, share);
			if (!silent) self.dispatch('share:add', list, email);
		},
		error: function (xhr) {
			if (xhr.status === 404) self.removeList(list);
			callback(xhr.err);
		}
	});
};

// Destroy share for list.
//
// @param		String			list			The list
// @param		String			email			The email to remove as share
// @param		Function		callback		The callback
// @param		Boolean			silent			True if no event should be dispatched
// @return		void
ETA.ShoppingList.prototype.destroyShare = function (list, email, callback, silent) {
	if (!this.hasUser) return;
	var self = this;
	this.api({
		type: 'delete',
		url: [this.getBaseURL(), list, 'shares', encodeURIComponent(email)].join('/'),
		success: function () {
			callback(null);
			if (!silent) self.dispatch('share:remove', list, email);
		},
		error: function (xhr) {
			if (xhr.status === 404) self.removeList(list);
			callback(xhr.err);
		}
	});
};

// Check if user is owner of list.
//
// @param		String			id				The list
// @param		String			email			The email of the owner
// @return		Boolean							True if user is owner
ETA.ShoppingList.prototype.isListOwner = function (id, email) {
	if (!this.hasUser) {
		return true;
	} else {
		var list = this.getList(id);

		// If there's no list to take ownership of.
		if (!list) return false;

		// If there isn't an owner, assume ownership.
		if (list.owner == null) return true;

		return list.owner.user === email;
	}
};

// Defaults
// -------------

ETA.ShoppingList.prototype.listDefaults = {
	name: null,
	access: 'private'
};

ETA.ShoppingList.prototype.itemDefaults = {
	description: null,
	count: 1,
	tick: false,
	offer_id: null
};

// Helpers
// -------------

// Get offer.
//
// @param		String			id				The offer id
// @return		Object							The offer attributes
ETA.ShoppingList.prototype.getOffer = function (id) {
	return this.offers[id];
};

// Get store.
//
// @param		String			id				The store id
// @return		Object							The store attributes
ETA.ShoppingList.prototype.getStore = function (id) {
	return this.stores[id];
};

// Poll lists.
//
// @return		void
ETA.ShoppingList.prototype.pollLists = function () {
	var self = this;
	this.clearListsTimeout();
	this.listsPollTimeout = setTimeout(function () {
		self.fetchLists(function () {
			self.pollLists();
		}, false);
	}, this.listsPollDelay);
};

// Poll items (only one list at a time).
//
// @return		void
ETA.ShoppingList.prototype.pollItems = function (list) {
	var self = this;
	this.clearItemsTimeout();
	this.itemsPollTimeout = setTimeout(function () {
		self.hasListBeenModified(list, function (err, hasBeenModified) {
			if (hasBeenModified) {
				self.fetchItems(list, function () {
					self.pollItems(list);
				}, false);
			} else {
				self.pollItems(list);
			}
		});
	}, this.itemsPollDelay);
};

// Clear all storage keys from namespace.
//
// @return		void
ETA.ShoppingList.prototype.removeStorage = function () {
	var keys = this.client.getStorageKeys(this.storagePrefix) || [];
	for (var i = 0, len = keys.length; i < len; i++) {
		this.client.removeStorage(keys[i]);
	}
};

// Get base URL.
//
// @return		String							The base URL
ETA.ShoppingList.prototype.getBaseURL = function () {
	if (!this.hasUser) {
		var error = new Error('ShoppingListError');
		error.name = 'UNAUTHORIZED_NO_USER';
		throw error;
	}
	return this.baseURL.replace('{user_id}', this.userId);
};

// Get modified date.
//
// @return		String							The ISO8601 date
ETA.ShoppingList.prototype.getModified = function () {
	return ETA.utils.toISOString(new Date());
};

// Clear lists poll timeout.
//
// @return		void
ETA.ShoppingList.prototype.clearListsTimeout = function () {
	clearTimeout(this.listsPollTimeout);
};

// Clear items poll timeout.
//
// @return		void
ETA.ShoppingList.prototype.clearItemsTimeout = function () {
	clearTimeout(this.itemsPollTimeout);
};

// Get an API promise.
//
// @return		Object							The promise object
ETA.ShoppingList.prototype.apiPromise = function () {
	var self = this;
	var token = ETA.utils.uuid();
	
	this.api(token);

	function done (options) {
		for (var i = 0, len = self.apiRequests.length; i < len; i++) {
			if (self.apiRequests[i] === token) {
				if (options) {
					self.apiRequests.splice(i, 1, options);
				} else {
					self.apiRequests.splice(i, 1);
				}
				break;
			}
		}
		self.apiRun();	
	}

	return {
		resolve: done,
		reject: done
	};
};

// API wrapper that ensures requests come in chronological order.
//
// @param		Object			options			The options
// @return		void
ETA.ShoppingList.prototype.api = function (options) {
	this.apiRequests.push(options);
	this.apiRun();
};

// API runner.
//
// @return		void
ETA.ShoppingList.prototype.apiRun = function () {
	if (this.apiRunning || this.apiRequests.length === 0) return;

	// If the element is a string, assume it's a promise token that will resolve itself.
	if (typeof this.apiRequests[0] === 'string') return;

	this.apiRunning = true;
	var self = this;
	var options = this.apiRequests.shift();
	var complete = options.complete;
	options.complete = function () {
		if (complete) complete.apply(this, arguments);
		self.apiRunning = false;
		self.apiRun();
	};
	this.client.api(options);
};

// Update user.
//
// @param		String			userId			The user id
// @return		void
ETA.ShoppingList.prototype.updateUser = function (userId) {
	this.userId = userId;
	this.hasUser = !!this.userId;
};
