// LocalStorage store with expiration logic.
ETA.Storage = {};

// List of storage ids.
ETA.Storage.ids = {};

// Check if `LocalStorage` is supported.
ETA.Storage.isSupported = ETA.features.localStorage;

// Get data for key with a fallback to a default value.
//
// @param		String			key				The key
// @param		Mixed			defaultValue	The default value if nothing was found for key
// @return		Mixed							The found value
ETA.Storage.get = function (key, defaultValue) {
	if (!ETA.Storage.isSupported || !key) return defaultValue;
	try {
		var now = new Date().getTime();
		var payload = window.localStorage.getItem(key);
		if (!payload) return defaultValue;
		payload = JSON.parse(payload);
		if (payload && payload.data && (!payload.expiration || now < payload.expiration)) {
			return payload.data;
		} else {
			return defaultValue;
		}
	} catch (err) {
		return defaultValue;
	}
};

// Set value for key with an optional lifetime.
//
// @param		String			key				The key
// @param		Mixed			data			The data
// @param		Number			lifetime		The lifetime in milliseconds
// @return		void
ETA.Storage.set = function (key, data, lifetime) {
	if (ETA.Storage.isSupported && key != null && data != null) {
		var id = ETA.utils.uuid();
		var payload = {
			data: data,
			expiration: null,
			id: id
		};

		// Save id.
		ETA.Storage.ids[id] = true;

		// Add current time to lifetime.
		if (!isNaN(lifetime) && lifetime > 0) {
			payload.expiration = new Date().getTime() + lifetime;
		}

		try {
			window.localStorage.setItem(key, JSON.stringify(payload));
		} catch (err) {
		}

		if (DEBUG) {
			console.log('ETA.Storage', 'Item set', key, payload);
		}
	}
};

// Remove data for key.
//
// @param		String			key				The key
// @return		void
ETA.Storage.remove = function (key) {
	if (ETA.Storage.isSupported) {
		try {
			window.localStorage.removeItem(key);
		} catch (err) {
		}
	}

	if (DEBUG) {
		console.log('ETA.Storage', 'Item removed', key);
	}
};

// Get all keys.
//
// @param		String			namespace		Optional namespace
// @return		Array							The saved keys
ETA.Storage.getKeys = function (namespace) {
	var keys = [];
	if (!ETA.Storage.isSupported) return keys;
	try {
		for (var i = 0, len = window.localStorage.length; i < len; i++) {
			var key = window.localStorage.key(i);
			if (namespace && key.indexOf(namespace) !== 0) continue;
			keys.push(key);
		}
	} catch (err) {
	}
	return keys;
};
