// Basic utilities where some are from the great underscore.js library:
// http://documentcloud.github.io/underscore/
ETA.utils = {};

// Create a function bound to a given object (assigning `this`, and arguments,
// optionally). Delegates to **ECMAScript 5**'s native `Function.bind` if
// available.
//
// @param		Function		func			The function
// @param		Object			context			The function context
// @return		Function						A function with the desired context
ETA.utils.bind = function (func, context) {
	var nativeBind = Function.prototype.bind;
	var slice = Array.prototype.slice;
	if (func.bind === nativeBind && nativeBind) return nativeBind.apply(func, slice.call(arguments, 1));
	var args = slice.call(arguments, 2);
	return function () {
		return func.apply(context, args.concat(slice.call(arguments)));
	};
};

// Is a given value an array?
//
// @param		Mixed			obj				The object to check
// @return		Boolean							True if object is an array.
ETA.utils.isArray = Array.isArray || function (obj) {
	return Object.prototype.toString.call(obj) === '[object Array]';
};

// Is a given variable an object?
//
// @param		Mixed			obj				The object to check
// @return		Boolean							True if object is an object
ETA.utils.isObject = function (obj) {
	return obj === Object(obj);
};

// Is a given variable a plain object?
//
// @param		Mixed			obj				The object to check
// @return		Boolean							True if object is a plain object
ETA.utils.isPlainObject = function (obj) {
	var ctor;
	var key;
	if (typeof obj !== 'object' || !obj || Object.prototype.toString.call(obj) != '[object Object]') {
		return false;
	}
	ctor = typeof obj.constructor == 'function' && obj.constructor.prototype;
	if (!ctor || !Object.prototype.hasOwnProperty.call(ctor, 'isPrototypeOf')) {
		return false;
	}
	for (key in obj) {}
	return key === void 0 || Object.prototype.hasOwnProperty.call(obj, key);
};

// Is a given variable a function?
//
// @param		Mixed			obj				The object to check
// @return		Boolean							True of object is a function
ETA.utils.isFunction = function (obj) {
	return !!(obj && obj.constructor && obj.call && obj.apply);
};

// Serialize a set of key/values into a query string.
//
// @param		Mixed			a				An array or object
// @return		String							A query string
ETA.utils.param = function (a) {
	var s = [];
	var r20 = /%20/g;
	var add = function (key, value) {
		// If value is a function, invoke it and return its value
		value = typeof value === 'function' ? value() : (value == null ? '' : value);
		s[s.length] = encodeURIComponent(key) + '=' + encodeURIComponent(value);
	};

	// If an array was passed in, assume that it is an array of form elements.
	if (ETA.utils.isArray(a)) {
		for (var i = 0, len = a.length; i < len; i++) {
			add(a[i].name, a[i].value);
		}
	} else if (ETA.utils.isPlainObject(a)) {
		for (var prefix in a) {
			ETA.utils.buildParams(prefix, a[prefix], add);
		}
	}

	// Return the resulting serialization.
	return s.join('&').replace(r20, '+');
};

// Build params.
//
// @param		String			prefix			The key prefix
// @param		Object			obj				The object
// @param		Function		add				The callback
ETA.utils.buildParams = function (prefix, obj, add) {
	if (ETA.utils.isArray(obj)) {
		for (var i = 0, len = obj.length; i < len; i++) {
			if (/\[\]$/.test(prefix)) {
				add(prefix, obj[i]);
			} else {
				ETA.utils.buildParams(prefix + '[' + (typeof obj[i] === 'object' ? i : '') + ']', obj[i], add);
			}
		}
	} else if (ETA.utils.isObject(obj)) {
		for (var name in obj) {
			ETA.utils.buildParams(prefix + '[' + name + ']', obj[name], add);
		}
	} else {
		add(prefix, obj);
	}
};

// Extend object.
//
// @return		Object							The extended object
ETA.utils.extend = function () {
	var options;
	var name;
	var src;
	var copy;
	var copyIsArray;
	var clone;
	var target = arguments[0] || {};
	var i = 1;
	var length = arguments.length;
	var deep = false;

	// Handle a deep copy situation.
	if (typeof target === 'boolean') {
		deep = target;
		target = arguments[1] || {};

		// Skip the boolean and the target.
		i = 2;
	}

	// Handle case when target is a string or something (possible in deep copy).
	if (typeof target !== 'object' && !ETA.utils.isFunction(target)) {
		target = {};
	}

	for (; i < length; i++) {
		// Only deal with non-null/undefined values.
		if ((options = arguments[i]) != null) {
			// Extend the base object.
			for (name in options) {
				src = target[name];
				copy = options[name];

				// Prevent never-ending loop.
				if (target === copy) {
					continue;
				}

				// Recurse if we're merging plain objects or arrays.
				if (deep && copy && (ETA.utils.isPlainObject(copy) || (copyIsArray = ETA.utils.isArray(copy)))) {
					if (copyIsArray) {
						copyIsArray = false;
						clone = src && ETA.utils.isArray(src) ? src : [];
					} else {
						clone = src && ETA.utils.isPlainObject(src) ? src : {};
					}

					// Never move original objects, clone them.
					target[name] = ETA.utils.extend(deep, clone, copy);

				// Don't bring in undefined values.
				} else if (copy !== undefined) {
					target[name] = copy;
				}
			}
		}
	}

	// Return the modified object
	return target;
};

// Convert degree to radian.
//
// @param		Number			deg				The degree
// @return		Number							The radian
ETA.utils.deg2rad = function (deg) {
	return deg * Math.PI / 180;
};

// Get distance between two points in metres:
// http://www.movable-type.co.uk/scripts/latlong.html
//
// @param		Number			latitude1		The first latitude
// @param		Number			longitude1		The first longitude
// @param		Number			latitude2		The second latitude
// @param		Number			longitude2		The second longitud
ETA.utils.getDistance = function (latitude1, longitude1, latitude2, longitude2) {
	var radius = 6371;
	var deltaLatitude = this.deg2rad(latitude2 - latitude1);
	var deltaLongitude = this.deg2rad(longitude2 - longitude1);
	
	latitude1 = this.deg2rad(latitude1);
	latitude2 = this.deg2rad(latitude2);

	var a = Math.sin(deltaLatitude / 2) * Math.sin(deltaLatitude / 2) + Math.sin(deltaLongitude / 2) * Math.sin(deltaLongitude / 2) * Math.cos(latitude1) * Math.cos(latitude2); 
	var c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a)); 
	var d = radius * c;

	return d * 1000;
};

// Clone an object.
//
// @param		Object			obj				The object
// @return		Object							The cloned object
ETA.utils.clone = function (obj) {
	if (obj == null || typeof(obj) != 'object') return obj;
	var temp = obj.constructor();

	for (var key in obj) {
		temp[key] = ETA.utils.clone(obj[key]);
	}

	return temp;
};

// Parse an ISO8601 string:
// https://github.com/csnover/js-iso8601/tree/lax
//
// @param		Object			date			A date object
// @return		Number							An epoch timestamp
ETA.utils.parseISOString = function (date) {
	var origParse = Date.parse;
	var numericKeys = [1, 4, 5, 6, 10, 11];
    var timestamp;
    var struct;
    var minutesOffset = 0;

    //              1 YYYY                 2 MM        3 DD              4 HH     5 mm        6 ss            7 msec         8 Z 9 ±    10 tzHH    11 tzmm
    if ((struct = /^(\d{4}|[+\-]\d{6})(?:-?(\d{2})(?:-?(\d{2}))?)?(?:[ T]?(\d{2}):?(\d{2})(?::?(\d{2})(?:[,\.](\d{1,}))?)?(?:(Z)|([+\-])(\d{2})(?::?(\d{2}))?)?)?$/.exec(date))) {
        // Avoid NaN timestamps caused by `undefined` values being passed to Date.UTC.
        for (var i = 0, k; (k = numericKeys[i]); ++i) {
            struct[k] = +struct[k] || 0;
        }

        // Allow undefined days and months.
        struct[2] = (+struct[2] || 1) - 1;
        struct[3] = +struct[3] || 1;

        // Allow arbitrary sub-second precision beyond milliseconds.
        struct[7] = struct[7] ? + (struct[7] + '00').substr(0, 3) : 0;

        // Timestamps without timezone identifiers should be considered local time.
        if ((struct[8] === undefined || struct[8] === '') && (struct[9] === undefined || struct[9] === '')) {
            timestamp = +new Date(struct[1], struct[2], struct[3], struct[4], struct[5], struct[6], struct[7]);
        } else {
            if (struct[8] !== 'Z' && struct[9] !== undefined) {
                minutesOffset = struct[10] * 60 + struct[11];

                if (struct[9] === '+') {
                    minutesOffset = 0 - minutesOffset;
                }
            }

            timestamp = Date.UTC(struct[1], struct[2], struct[3], struct[4], struct[5] + minutesOffset, struct[6], struct[7]);
        }
    } else {
        timestamp = origParse ? origParse(date) : NaN;
    }

    return timestamp;
};

// Convert to ISO8601 string.
//
// @param		Object			date			A date object
// @return		String							The ISO8601 string
ETA.utils.toISOString = function (date) {
	function f (n) {
		return n < 10 ? '0' + n : n;
	}
		
	if (Date.prototype.toISOString) {
		return date.toISOString();
	} else {
        return date.getUTCFullYear()  + '-' +
            f(date.getUTCMonth() + 1) + '-' +
            f(date.getUTCDate())      + 'T' +
            f(date.getUTCHours())     + ':' +
            f(date.getUTCMinutes())   + ':' +
            f(date.getUTCSeconds())   + 'Z';
	}
};

// Generate UUID:
// http://www.broofa.com/Tools/Math.uuid.htm
//
// @return		String						The UUID string
ETA.utils.uuid = function () {
	var chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'.split('');
	var uuid = [];
	var radix = chars.length;

	// rfc4122, version 4 form.
	var r;

	// rfc4122 requires these characters.
	uuid[8] = uuid[13] = uuid[18] = uuid[23] = '-';
	uuid[14] = '4';

	// Fill in random data.  At i==19 set the high bits of clock sequence as
	// per rfc4122, sec. 4.1.5.
	for (var i = 0; i < 36; i++) {
		if (!uuid[i]) {
			r = 0 | Math.random()*16;
			uuid[i] = chars[(i === 19) ? (r & 0x3) | 0x8 : r];
		}
	}

	return uuid.join('').toLowerCase();
};

// SHA256 crypto.
//
// @param		String			s			The string to do a SHA256 of
// @return		String						The SHA256 string
ETA.utils.SHA256 = function (s) {
	var chrsz = 8;
	var hexcase = 0;
 
	function safe_add (x, y) {
		var lsw = (x & 0xFFFF) + (y & 0xFFFF);
		var msw = (x >> 16) + (y >> 16) + (lsw >> 16);
		return (msw << 16) | (lsw & 0xFFFF);
	}
 
	function S (X, n) { return ( X >>> n ) | (X << (32 - n)); }
	function R (X, n) { return ( X >>> n ); }
	function Ch (x, y, z) { return ((x & y) ^ ((~x) & z)); }
	function Maj (x, y, z) { return ((x & y) ^ (x & z) ^ (y & z)); }
	function Sigma0256 (x) { return (S(x, 2) ^ S(x, 13) ^ S(x, 22)); }
	function Sigma1256 (x) { return (S(x, 6) ^ S(x, 11) ^ S(x, 25)); }
	function Gamma0256 (x) { return (S(x, 7) ^ S(x, 18) ^ R(x, 3)); }
	function Gamma1256 (x) { return (S(x, 17) ^ S(x, 19) ^ R(x, 10)); }
 
	function core_sha256 (m, l) {
		var K = new Array(0x428A2F98, 0x71374491, 0xB5C0FBCF, 0xE9B5DBA5, 0x3956C25B, 0x59F111F1, 0x923F82A4, 0xAB1C5ED5, 0xD807AA98, 0x12835B01, 0x243185BE, 0x550C7DC3, 0x72BE5D74, 0x80DEB1FE, 0x9BDC06A7, 0xC19BF174, 0xE49B69C1, 0xEFBE4786, 0xFC19DC6, 0x240CA1CC, 0x2DE92C6F, 0x4A7484AA, 0x5CB0A9DC, 0x76F988DA, 0x983E5152, 0xA831C66D, 0xB00327C8, 0xBF597FC7, 0xC6E00BF3, 0xD5A79147, 0x6CA6351, 0x14292967, 0x27B70A85, 0x2E1B2138, 0x4D2C6DFC, 0x53380D13, 0x650A7354, 0x766A0ABB, 0x81C2C92E, 0x92722C85, 0xA2BFE8A1, 0xA81A664B, 0xC24B8B70, 0xC76C51A3, 0xD192E819, 0xD6990624, 0xF40E3585, 0x106AA070, 0x19A4C116, 0x1E376C08, 0x2748774C, 0x34B0BCB5, 0x391C0CB3, 0x4ED8AA4A, 0x5B9CCA4F, 0x682E6FF3, 0x748F82EE, 0x78A5636F, 0x84C87814, 0x8CC70208, 0x90BEFFFA, 0xA4506CEB, 0xBEF9A3F7, 0xC67178F2);
		var HASH = new Array(0x6A09E667, 0xBB67AE85, 0x3C6EF372, 0xA54FF53A, 0x510E527F, 0x9B05688C, 0x1F83D9AB, 0x5BE0CD19);
		var W = new Array(64);
		var a, b, c, d, e, f, g, h, i, j;
		var T1, T2;
 
		m[l >> 5] |= 0x80 << (24 - l % 32);
		m[((l + 64 >> 9) << 4) + 15] = l;
 
		for (i = 0; i < m.length; i+=16 ) {
			a = HASH[0];
			b = HASH[1];
			c = HASH[2];
			d = HASH[3];
			e = HASH[4];
			f = HASH[5];
			g = HASH[6];
			h = HASH[7];
 
			for (j = 0; j < 64; j++) {
				if (j < 16) W[j] = m[j + i];
				else W[j] = safe_add(safe_add(safe_add(Gamma1256(W[j - 2]), W[j - 7]), Gamma0256(W[j - 15])), W[j - 16]);
 
				T1 = safe_add(safe_add(safe_add(safe_add(h, Sigma1256(e)), Ch(e, f, g)), K[j]), W[j]);
				T2 = safe_add(Sigma0256(a), Maj(a, b, c));
 
				h = g;
				g = f;
				f = e;
				e = safe_add(d, T1);
				d = c;
				c = b;
				b = a;
				a = safe_add(T1, T2);
			}
 
			HASH[0] = safe_add(a, HASH[0]);
			HASH[1] = safe_add(b, HASH[1]);
			HASH[2] = safe_add(c, HASH[2]);
			HASH[3] = safe_add(d, HASH[3]);
			HASH[4] = safe_add(e, HASH[4]);
			HASH[5] = safe_add(f, HASH[5]);
			HASH[6] = safe_add(g, HASH[6]);
			HASH[7] = safe_add(h, HASH[7]);
		}
		return HASH;
	}
 
	function str2binb (str) {
		var bin = Array();
		var mask = (1 << chrsz) - 1;
		for(var i = 0; i < str.length * chrsz; i += chrsz) {
			bin[i>>5] |= (str.charCodeAt(i / chrsz) & mask) << (24 - i%32);
		}
		return bin;
	}
 
	function Utf8Encode (string) {
		string = string.replace(/\r\n/g, '\n');
		var utftext = '';
 
		for (var n = 0; n < string.length; n++) {
			var c = string.charCodeAt(n);
 
			if (c < 128) {
				utftext += String.fromCharCode(c);
			} else if((c > 127) && (c < 2048)) {
				utftext += String.fromCharCode((c >> 6) | 192);
				utftext += String.fromCharCode((c & 63) | 128);
			} else {
				utftext += String.fromCharCode((c >> 12) | 224);
				utftext += String.fromCharCode(((c >> 6) & 63) | 128);
				utftext += String.fromCharCode((c & 63) | 128);
			}
		}
 
		return utftext;
	}
 
	function binb2hex (binarray) {
		var hex_tab = hexcase ? '0123456789ABCDEF' : '0123456789abcdef';
		var str = '';
		for(var i = 0; i < binarray.length * 4; i++) {
			str += hex_tab.charAt((binarray[i>>2] >> ((3 - i%4)*8+4)) & 0xF) +
			hex_tab.charAt((binarray[i>>2] >> ((3 - i%4)*8  )) & 0xF);
		}
		return str;
	}
 
	s = Utf8Encode(s);
	return binb2hex(core_sha256(str2binb(s), s.length * chrsz));
};

// Extract target URL.
//
// @param		String			url				The URL to extract the target from.
// @return		String							The target URL
ETA.utils.getTarget = function (url) {
	var parser = document.createElement('a');
	parser.href = url;
	return parser.protocol + '//' + parser.host;
};

// An empty function.
//
// @return		void
ETA.utils.emptyFunction = function () {};

// Get URL parameter.
//
// @param		String			name			The parameter name
// @param		String			href			The location href
// @return		String							The parameter value
ETA.utils.getParameter = function (name, href) {
	var results = new RegExp('[\\?&]' + name + '=([^&#]*)').exec(href || window.location.href);
	return results && results[1];
};

// Update URL parameter.
//
// @param		String			url				The URL to update
// @param		String			key				The parameter key
// @param		String			value			The parameter value
// @return		String							The updated URL
ETA.utils.updateParameter = function (url, key, value) {
	var re = new RegExp('([?|&])' + key + '=.*?(&|$)', 'i');
	var separator = url.indexOf('?') !== -1 ? '&' : '?';
	key = encodeURIComponent(key);
	value = encodeURIComponent(value);
	if (url.match(re)) {
		return url.replace(re, '$1' + key + '=' + value + '$2');
	} else {
		return url + separator + key + '=' + value;
	}
};

// Encode string to base64.
ETA.utils.base64 = (function () {
    var b64chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
    var b64tab = function (bin) {
        var t = {};
        for (var i = 0, l = bin.length; i < l; i++) t[bin.charAt(i)] = i;
        return t;
    }(b64chars);
    var fromCharCode = String.fromCharCode;
    var cc;
    var cb_utob = function (c) {
        if (c.length < 2) {
            cc = c.charCodeAt(0);
            return cc < 0x80 ? c
                : cc < 0x800 ? (fromCharCode(0xc0 | (cc >>> 6)) + fromCharCode(0x80 | (cc & 0x3f)))
                : (fromCharCode(0xe0 | ((cc >>> 12) & 0x0f)) + fromCharCode(0x80 | ((cc >>>  6) & 0x3f)) + fromCharCode(0x80 | ( cc         & 0x3f)));
        } else {
            cc = 0x10000 + (c.charCodeAt(0) - 0xD800) * 0x400 + (c.charCodeAt(1) - 0xDC00);
            return (fromCharCode(0xf0 | ((cc >>> 18) & 0x07)) + fromCharCode(0x80 | ((cc >>> 12) & 0x3f)) + fromCharCode(0x80 | ((cc >>>  6) & 0x3f)) + fromCharCode(0x80 | ( cc         & 0x3f)));
        }
    };
    var re_utob = /[\uD800-\uDBFF][\uDC00-\uDFFFF]|[^\x00-\x7F]/g;
    var utob = function (u) {
        return u.replace(re_utob, cb_utob);
    };
    var cb_encode = function (ccc) {
        var padlen = [0, 2, 1][ccc.length % 3],
        ord = ccc.charCodeAt(0) << 16
            | ((ccc.length > 1 ? ccc.charCodeAt(1) : 0) << 8)
            | ((ccc.length > 2 ? ccc.charCodeAt(2) : 0)),
        chars = [
            b64chars.charAt( ord >>> 18),
            b64chars.charAt((ord >>> 12) & 63),
            padlen >= 2 ? '=' : b64chars.charAt((ord >>> 6) & 63),
            padlen >= 1 ? '=' : b64chars.charAt(ord & 63)
        ];
        return chars.join('');
    };
    var btoa = window.btoa || function (b) {
        return b.replace(/[\s\S]{1,3}/g, cb_encode);
    };
    var _encode = function (u) { return btoa(utob(u)); };
    var encode = function (u, urisafe) {
        return !urisafe ? _encode(u) : _encode(u).replace(/[+\/]/g, function (m0) {
            return m0 == '+' ? '-' : '_';
        }).replace(/=/g, '');
    };
    var encodeURI = function (u) { return encode(u, true); };
    // decoder stuff
    var re_btou = new RegExp([
        '[\xC0-\xDF][\x80-\xBF]',
        '[\xE0-\xEF][\x80-\xBF]{2}',
        '[\xF0-\xF7][\x80-\xBF]{3}'
    ].join('|'), 'g');
    var cb_btou = function (cccc) {
        switch(cccc.length) {
        case 4:
            var cp = ((0x07 & cccc.charCodeAt(0)) << 18)
                |    ((0x3f & cccc.charCodeAt(1)) << 12)
                |    ((0x3f & cccc.charCodeAt(2)) <<  6)
                |     (0x3f & cccc.charCodeAt(3)),
            offset = cp - 0x10000;
            return (fromCharCode((offset  >>> 10) + 0xD800) + fromCharCode((offset & 0x3FF) + 0xDC00));
        case 3:
            return fromCharCode(
                ((0x0f & cccc.charCodeAt(0)) << 12)
                    | ((0x3f & cccc.charCodeAt(1)) << 6)
                    |  (0x3f & cccc.charCodeAt(2))
            );
        default:
            return  fromCharCode(
                ((0x1f & cccc.charCodeAt(0)) << 6)
                    |  (0x3f & cccc.charCodeAt(1))
            );
        }
    };
    var btou = function (b) {
        return b.replace(re_btou, cb_btou);
    };
    var cb_decode = function (cccc) {
        var len = cccc.length,
        padlen = len % 4,
        n = (len > 0 ? b64tab[cccc.charAt(0)] << 18 : 0)
            | (len > 1 ? b64tab[cccc.charAt(1)] << 12 : 0)
            | (len > 2 ? b64tab[cccc.charAt(2)] <<  6 : 0)
            | (len > 3 ? b64tab[cccc.charAt(3)]       : 0),
        chars = [
            fromCharCode( n >>> 16),
            fromCharCode((n >>>  8) & 0xff),
            fromCharCode( n         & 0xff)
        ];
        chars.length -= [0, 0, 2, 1][padlen];
        return chars.join('');
    };
    var atob = window.atob || function (a){
        return a.replace(/[\s\S]{1,4}/g, cb_decode);
    };
    var _decode = function (a) { return btou(atob(a)); };
    var decode = function (a){
        return _decode(
            a.replace(/[-_]/g, function (m0) { return m0 == '-' ? '+' : '/'; })
                .replace(/[^A-Za-z0-9\+\/]/g, '')
        );
    };

    return {
        atob: atob,
        btoa: btoa,
        fromBase64: decode,
        toBase64: encode,
        utob: utob,
        encode: encode,
        encodeURI: encodeURI,
        btou: btou,
        decode: decode
    };
})();
